import builtins
import os
import pathlib
import re
import types
from unittest.mock import AsyncMock, Mock, PropertyMock, call

import pytest

from upsies import errors, utils
from upsies.trackers.base import TrackerJobsBase


@pytest.fixture
def tracker():
    tracker = Mock()
    tracker.configure_mock(
        name='AsdF',
        options={
            'announce' : 'http://foo.bar',
            'source'   : 'AsdF',
            'exclude'  : ('a', 'b'),
        },
    )
    return tracker


def make_TestTrackerJobs(**kwargs):
    class TestTrackerJobs(TrackerJobsBase):
        jobs_before_upload = PropertyMock()

    default_kwargs = {
        'content_path': '',
        'tracker': Mock(
            options={
                'announce' : 'http://foo.bar',
                'source'   : 'AsdF',
                'exclude'  : ('a', 'b'),
            },
        ),
        'reuse_torrent_path': '',
        'btclient': Mock(),
        'torrent_destination': '',
        'image_host': '',
        'screenshots_optimization': 'my optimization level',
        'common_job_args': {},
    }
    return TestTrackerJobs(**{**default_kwargs, **kwargs})


def test_arguments():
    kwargs = {
        'content_path': Mock(),
        'tracker': Mock(),
        'reuse_torrent_path': Mock(),
        'exclude_files': Mock(),
        'btclient': Mock(),
        'torrent_destination': Mock(),
        'image_host': Mock(),
        'options': {'mock': 'config'},
    }
    tracker_jobs = make_TestTrackerJobs(**kwargs)
    for k, v in kwargs.items():
        assert getattr(tracker_jobs, k) is v


@pytest.mark.parametrize(
    argnames='common_job_args, overload, exp_return_value',
    argvalues=(
        (
            {'home_directory': 'initial/path', 'ignore_cache': 'mock bool'},
            {'home_directory': 'overloaded/path'},
            {'home_directory': 'overloaded/path', 'ignore_cache': 'mock bool'},
        ),
        (
            {'home_directory': 'initial/path', 'ignore_cache': 'mock bool'},
            {'additional_value': 'something'},
            {'home_directory': 'initial/path', 'ignore_cache': 'mock bool', 'additional_value': 'something'},
        ),
    ),
    ids=lambda v: repr(v),
)
def test_common_job_args(common_job_args, overload, exp_return_value):
    tracker_jobs = make_TestTrackerJobs(common_job_args=common_job_args)
    return_value = tracker_jobs.common_job_args(**overload)
    assert return_value == exp_return_value


def test_signals(mocker):
    Signal_mock = mocker.patch('upsies.utils.signal.Signal')
    tracker_jobs = make_TestTrackerJobs()
    assert tracker_jobs.signal is Signal_mock.return_value
    assert Signal_mock.call_args_list == [call('warning', 'error', 'exception')]
    tracker_jobs.warn('foo')
    assert Signal_mock.return_value.emit.call_args_list == [
        call('warning', 'foo'),
    ]
    tracker_jobs.error('bar')
    assert Signal_mock.return_value.emit.call_args_list == [
        call('warning', 'foo'),
        call('error', 'bar'),
    ]
    tracker_jobs.exception('baz')
    assert Signal_mock.return_value.emit.call_args_list == [
        call('warning', 'foo'),
        call('error', 'bar'),
        call('exception', 'baz'),
    ]


def test_imdb():
    tracker_jobs = make_TestTrackerJobs()
    assert isinstance(tracker_jobs.imdb, utils.webdbs.imdb.ImdbApi)

def test_tmdb():
    tracker_jobs = make_TestTrackerJobs()
    assert isinstance(tracker_jobs.tvmaze, utils.webdbs.tvmaze.TvmazeApi)

def test_tvmaze():
    tracker_jobs = make_TestTrackerJobs()
    assert isinstance(tracker_jobs.tvmaze, utils.webdbs.tvmaze.TvmazeApi)


def test_jobs_after_upload(mocker):
    add_torrent_job_mock = mocker.patch('upsies.trackers.base.TrackerJobsBase.add_torrent_job')
    copy_torrent_job_mock = mocker.patch('upsies.trackers.base.TrackerJobsBase.copy_torrent_job')
    tracker_jobs = make_TestTrackerJobs()
    assert tracker_jobs.jobs_after_upload == (
        add_torrent_job_mock,
        copy_torrent_job_mock,
    )


def test_isolated_jobs(mocker):
    tracker_jobs = make_TestTrackerJobs()
    assert tracker_jobs.isolated_jobs == ()


class Job(str):
    def __new__(cls, name, *, prejobs=(), presignals={}):
        self = super().__new__(
            cls,
            f'Job({name!r}, '
            + 'prejobs=' + ', '.join(repr(j.name) for j in prejobs) + ', '
            + 'presignals=' + ', '.join(repr(j.name) for j in presignals) + ', '
        )
        self.name = str(name)
        self.prejobs = tuple(prejobs)
        self.presignals = dict(presignals)
        return self

@pytest.mark.parametrize(
    argnames='jobs, exp_job_names',
    argvalues=(
        pytest.param(
            [
                Job(name='a'),
                Job(name='b'),
                Job(name='c'),
            ],
            ['a', 'b', 'c'],
            id='No dependencies',
        ),

        # NOTE: `...` or `Ellipsis` is a built-in Python singleton. We use it
        #       here to show that we only need the keys/jobs in `presignals` and
        #       the values are ignored.

        pytest.param(
            [
                Job(name='a', prejobs=[
                    Job(name='x', prejobs=[Job(name='1')]),
                    Job(name='y', prejobs=[Job(name='2', prejobs=[Job(name='ɑ')])]),
                    Job(name='c', prejobs=[]),
                ]),
                Job(name='b', prejobs=[
                    Job(name='x', prejobs=[Job(name='1')]),
                    Job(name='z', prejobs=[Job(name='3')]),
                ]),
                Job(name='c', prejobs=[]),

            ],
            ['a', 'x', '1', 'y', '2', 'ɑ', 'c', 'b', 'z', '3'],
            id='Dependencies from prejobs',
        ),

        pytest.param(
            [
                Job(name='a', presignals={Job(name='ɑ'): ...}),
                Job(name='b', presignals={
                    Job(name='x'): ...,
                    Job(name='y', presignals={Job(name='1'): ...}): ...,
                }),
                Job(name='c', presignals={
                    Job(name='y', presignals={Job(name='1'): ...}): ...,
                    Job(name='z', presignals={
                        Job(name='2'): ...,
                        Job(name='3', presignals={Job(name='ɑ'): ...}): ...,
                    }): ...,
                }),
            ],
            ['a', 'ɑ', 'b', 'x', 'y', '1', 'c', 'z', '2', '3'],
            id='Dependencies from presignals',
        ),

        pytest.param(
            [
                Job(name='c', prejobs=[
                    Job(name='z', presignals={
                        Job(name='2'): ...,
                    }),
                ]),
            ],
            ['c', 'z', '2'],
            id='Job depends on prejobs with presignals',
        ),

        pytest.param(
            [
                Job(name='c', presignals={
                    Job(name='z', prejobs=[
                        Job(name='2'),
                    ]): ...,
                }),
            ],
            ['c', 'z', '2'],
            id='Job depends on presignals with prejobs',
        ),

        pytest.param(
            [
                Job(name='a', prejobs=[Job(name='ɑ')], presignals={Job(name='1'): ...}),
                Job(name='b', presignals={
                    Job(name='x'): ...,
                    Job(name='y', prejobs=[Job(name='1')]): ...,
                }),
                Job(name='c', prejobs=[
                    Job(name='y', prejobs=[Job(name='1')]),
                    Job(
                        name='z',
                        prejobs=[Job(name='10'), Job(name='20')],
                        presignals={
                            Job(
                                name='2',
                                presignals={
                                    Job(
                                        name='β',
                                        prejobs=[
                                            Job(name='a', prejobs=[Job(name='ɑ')], presignals={Job(name='1'): ...}),
                                            Job(name='y', prejobs=[Job(name='1')]),
                                        ],
                                    ): ...,
                                },
                            ): ...,
                            Job(
                                name='3',
                                prejobs=[
                                    Job(name='a', prejobs=[Job(name='ɑ')], presignals={Job(name='1'): ...}),
                                ],
                                presignals={
                                    Job(name='a', prejobs=[Job(name='ɑ')], presignals={Job(name='1'): ...}): ...,
                                    Job(name='y', prejobs=[Job(name='1')]): ...,
                                },
                            ): ...,
                        },
                    ),
                ]),
            ],
            ['a', 'ɑ', '1', 'b', 'x', 'y', 'c', 'z', '10', '20', '2', 'β', '3'],
            id='Lots of mixed nested dependencies',
        ),
    ),
)
def test_get_job_and_dependencies(jobs, exp_job_names):
    tracker_jobs = make_TestTrackerJobs()
    required_jobs = tracker_jobs.get_job_and_dependencies(*jobs)
    assert [j.name for j in required_jobs] == exp_job_names


@pytest.mark.parametrize('isolated_jobs', ((), ('mock isolated job',)))
@pytest.mark.parametrize(
    argnames='jobs_before_upload, exp_submission_ok',
    argvalues=(
        ((), False),
        # All jobs exited successfully
        ((Mock(exit_code=0, is_enabled=True), Mock(exit_code=0, is_enabled=True), Mock(exit_code=0, is_enabled=True)), True),

        # Job exited with non-zero exit_code
        ((Mock(exit_code=0, is_enabled=True), Mock(exit_code=0, is_enabled=True), Mock(exit_code=1, is_enabled=True)), False),
        ((Mock(exit_code=0, is_enabled=True), Mock(exit_code=1, is_enabled=True), Mock(exit_code=0, is_enabled=True)), False),
        ((Mock(exit_code=1, is_enabled=True), Mock(exit_code=0, is_enabled=True), Mock(exit_code=0, is_enabled=True)), False),

        # Job is disabled
        ((Mock(exit_code=None, is_enabled=False), Mock(exit_code=0, is_enabled=True), Mock(exit_code=0, is_enabled=True)), True),
        ((Mock(exit_code=0, is_enabled=True), Mock(exit_code=None, is_enabled=False), Mock(exit_code=0, is_enabled=True)), True),
        ((Mock(exit_code=0, is_enabled=True), Mock(exit_code=0, is_enabled=False), Mock(exit_code=None, is_enabled=False)), True),
    ),
)
def test_submission_ok(isolated_jobs, jobs_before_upload, exp_submission_ok, mocker):
    tracker_jobs = make_TestTrackerJobs()
    mocker.patch.object(type(tracker_jobs), 'jobs_before_upload', PropertyMock(return_value=jobs_before_upload))
    mocker.patch.object(type(tracker_jobs), 'isolated_jobs', PropertyMock(return_value=isolated_jobs))
    if isolated_jobs:
        assert tracker_jobs.submission_ok is False
    else:
        assert tracker_jobs.submission_ok == exp_submission_ok


@pytest.mark.parametrize(
    argnames='job_name, tracker_name, exp_name',
    argvalues=(
        ('myjob', 'mytracker', 'myjob.mytracker'),
        ('myjob.mytracker', 'mytracker', 'myjob.mytracker'),
    ),
)
def test_get_job_name(job_name, tracker_name, exp_name, tracker, mocker):
    tracker.name = tracker_name
    tracker_jobs = make_TestTrackerJobs(
        content_path='path/to/content',
        tracker=tracker,
        common_job_args={'home_directory': 'path/to/home', 'ignore_cache': 'mock bool'},
    )
    assert tracker_jobs.get_job_name(job_name) == exp_name


def test_create_torrent_job(mocker):
    CreateTorrentJob_mock = mocker.patch('upsies.jobs.torrent.CreateTorrentJob')
    tracker_jobs = make_TestTrackerJobs(
        content_path='path/to/content',
        reuse_torrent_path='path/to/existing.torrent',
        tracker='mock tracker',
        exclude_files=('a', 'b', 'c'),
        common_job_args={'home_directory': 'path/to/home', 'ignore_cache': 'mock bool'},
    )
    mocker.patch.object(tracker_jobs, 'make_precondition')
    assert tracker_jobs.create_torrent_job is CreateTorrentJob_mock.return_value
    assert CreateTorrentJob_mock.call_args_list == [
        call(
            content_path='path/to/content',
            reuse_torrent_path='path/to/existing.torrent',
            tracker='mock tracker',
            exclude_files=('a', 'b', 'c'),
            precondition=tracker_jobs.make_precondition.return_value,
            home_directory='path/to/home',
            ignore_cache='mock bool',
        ),
    ]
    assert tracker_jobs.make_precondition.call_args_list == [call('create_torrent_job')]

def test_create_torrent_job_is_singleton(mocker):
    mocker.patch('upsies.jobs.torrent.CreateTorrentJob', side_effect=(Mock(), Mock()))
    tracker_jobs = make_TestTrackerJobs(
        content_path='path/to/content',
        tracker='mock tracker',
        common_job_args={'home_directory': 'path/to/home', 'ignore_cache': 'mock bool'},
    )
    assert tracker_jobs.create_torrent_job is tracker_jobs.create_torrent_job


@pytest.mark.parametrize(
    argnames='btclient, create_torrent_job, exp_add_torrent_job_is_None',
    argvalues=(
        (Mock(), Mock(), False),
        (None, Mock(), True),
        (Mock(), None, True),
    ),
)
def test_add_torrent_job(btclient, create_torrent_job, exp_add_torrent_job_is_None, mocker):
    AddTorrentJob_mock = mocker.patch('upsies.jobs.torrent.AddTorrentJob')
    tracker_jobs = make_TestTrackerJobs(
        content_path='path/to/content',
        tracker='mock tracker',
        common_job_args={'home_directory': 'path/to/home', 'ignore_cache': 'mock bool'},
        btclient=btclient,
    )
    mocker.patch.object(type(tracker_jobs), 'create_torrent_job', PropertyMock(return_value=create_torrent_job))
    mocker.patch.object(tracker_jobs, 'make_precondition')
    if exp_add_torrent_job_is_None:
        assert tracker_jobs.add_torrent_job is None
        assert AddTorrentJob_mock.call_args_list == []
        assert tracker_jobs.make_precondition.call_args_list == []
    else:
        assert tracker_jobs.add_torrent_job is AddTorrentJob_mock.return_value
        assert AddTorrentJob_mock.call_args_list == [
            call(
                autostart=False,
                btclient=btclient,
                precondition=tracker_jobs.make_precondition.return_value,
                home_directory='path/to/home',
                ignore_cache='mock bool',
            ),
        ]
        assert tracker_jobs.create_torrent_job.signal.register.call_args_list == [
            call('output', tracker_jobs.add_torrent_job.enqueue),
            call('finished', tracker_jobs.finalize_add_torrent_job),
        ]
        assert tracker_jobs.make_precondition.call_args_list == [call('add_torrent_job')]

def test_add_torrent_job_is_singleton(mocker):
    mocker.patch('upsies.jobs.torrent.AddTorrentJob', side_effect=(Mock(), Mock()))
    mocker.patch('upsies.jobs.torrent.CreateTorrentJob')
    tracker_jobs = make_TestTrackerJobs(
        content_path='path/to/content',
        tracker='mock tracker',
        common_job_args={'home_directory': 'path/to/home', 'ignore_cache': 'mock bool'},
        btclient='bittorrent client mock',
    )
    assert tracker_jobs.add_torrent_job is tracker_jobs.add_torrent_job


def test_finalize_add_torrent_job(mocker):
    mocker.patch('upsies.jobs.torrent.AddTorrentJob')
    mocker.patch('upsies.jobs.torrent.CreateTorrentJob')
    tracker_jobs = make_TestTrackerJobs(
        content_path='path/to/content',
        tracker='mock tracker',
        common_job_args={'home_directory': 'path/to/home', 'ignore_cache': 'mock bool'},
        btclient='bittorrent client mock',
    )
    mocker.patch.object(tracker_jobs, 'add_torrent_job')
    tracker_jobs.finalize_add_torrent_job('mock job')
    assert tracker_jobs.add_torrent_job.finalize.call_args_list == [call()]


@pytest.mark.parametrize(
    argnames='torrent_destination, create_torrent_job, exp_copy_torrent_job_is_None',
    argvalues=(
        ('some/path', Mock(), False),
        (None, Mock(), True),
        ('some/path', None, True),
    ),
)
def test_copy_torrent_job(torrent_destination, create_torrent_job, exp_copy_torrent_job_is_None, tracker, mocker):
    CopyTorrentJob_mock = mocker.patch('upsies.jobs.torrent.CopyTorrentJob')
    tracker_jobs = make_TestTrackerJobs(
        content_path='path/to/content',
        common_job_args={'home_directory': 'path/to/home', 'ignore_cache': 'mock bool'},
        torrent_destination=torrent_destination,
        tracker=tracker,
    )
    mocker.patch.object(type(tracker_jobs), 'create_torrent_job', PropertyMock(return_value=create_torrent_job))
    mocker.patch.object(tracker_jobs, 'make_precondition')
    if exp_copy_torrent_job_is_None:
        assert tracker_jobs.copy_torrent_job is None
        assert CopyTorrentJob_mock.call_args_list == []
        assert tracker_jobs.make_precondition.call_args_list == []
    else:
        assert tracker_jobs.copy_torrent_job is CopyTorrentJob_mock.return_value
        assert CopyTorrentJob_mock.call_args_list == [
            call(
                autostart=False,
                destination=torrent_destination,
                precondition=tracker_jobs.make_precondition.return_value,
                home_directory='path/to/home',
                ignore_cache='mock bool',
            ),
        ]
        assert tracker_jobs.create_torrent_job.signal.register.call_args_list == [
            call('output', tracker_jobs.copy_torrent_job.enqueue),
            call('finished', tracker_jobs.finalize_copy_torrent_job),
        ]
        assert tracker_jobs.make_precondition.call_args_list == [call('copy_torrent_job')]

def test_copy_torrent_job_is_singleton(mocker):
    mocker.patch('upsies.jobs.torrent.CopyTorrentJob', side_effect=(Mock(), Mock()))
    mocker.patch('upsies.jobs.torrent.CreateTorrentJob')
    tracker_jobs = make_TestTrackerJobs(
        content_path='path/to/content',
        common_job_args={'home_directory': 'path/to/home', 'ignore_cache': 'mock bool'},
        torrent_destination='path/to/torrent/destination',
    )
    assert tracker_jobs.copy_torrent_job is tracker_jobs.copy_torrent_job


def test_finalize_copy_torrent_job(mocker):
    mocker.patch('upsies.jobs.torrent.CopyTorrentJob')
    mocker.patch('upsies.jobs.torrent.CreateTorrentJob')
    tracker_jobs = make_TestTrackerJobs(
        content_path='path/to/content',
        tracker='mock tracker',
        common_job_args={'home_directory': 'path/to/home', 'ignore_cache': 'mock bool'},
        btclient='bittorrent client mock',
    )
    mocker.patch.object(tracker_jobs, 'copy_torrent_job')
    tracker_jobs.finalize_copy_torrent_job('mock job')
    assert tracker_jobs.copy_torrent_job.finalize.call_args_list == [call()]


def test_release_name(mocker):
    ReleaseName_mock = mocker.patch('upsies.utils.release.ReleaseName')
    tracker_jobs = make_TestTrackerJobs(
        content_path='path/to/content',
        common_job_args={'home_directory': 'path/to/home', 'ignore_cache': 'mock bool'},
    )
    assert tracker_jobs.release_name is ReleaseName_mock.return_value
    assert ReleaseName_mock.call_args_list == [
        call(
            path='path/to/content',
            translate=tracker_jobs.release_name_translation,
            separator=tracker_jobs.release_name_separator,
        ),
    ]


def test_release_name_job(tracker, mocker):
    TextFieldJob_mock = mocker.patch('upsies.jobs.dialog.TextFieldJob')
    tracker_jobs = make_TestTrackerJobs(
        content_path='path/to/content',
        tracker=tracker,
        common_job_args={'home_directory': 'path/to/home', 'ignore_cache': 'mock bool'},
    )

    mocker.patch.object(type(tracker_jobs), 'release_name', PropertyMock())
    mocker.patch.object(tracker_jobs, 'get_job_name')
    mocker.patch.object(tracker_jobs, 'make_precondition')

    release_name_job = tracker_jobs.release_name_job
    assert release_name_job is TextFieldJob_mock.return_value
    assert TextFieldJob_mock.call_args_list == [
        call(
            name=tracker_jobs.get_job_name.return_value,
            label='Release Name',
            text=str(tracker_jobs.release_name),
            callbacks={
                'output': tracker_jobs.release_name.set_release_info,
            },
            precondition=tracker_jobs.make_precondition.return_value,
            home_directory='path/to/home',
            ignore_cache='mock bool',
        ),
    ]
    assert tracker_jobs.make_precondition.call_args_list == [call('release_name_job')]

def test_release_name_job_is_singleton(mocker):
    mocker.patch('upsies.jobs.dialog.TextFieldJob', side_effect=(Mock(), Mock()))
    tracker_jobs = make_TestTrackerJobs(
        content_path='path/to/content',
        common_job_args={'home_directory': 'path/to/home', 'ignore_cache': 'mock bool'},
    )
    assert tracker_jobs.release_name_job is tracker_jobs.release_name_job


@pytest.mark.asyncio
async def test_update_release_name_from(mocker):
    tracker_jobs = make_TestTrackerJobs(
        content_path='path/to/content',
        common_job_args={'home_directory': 'path/to/home', 'ignore_cache': 'mock bool'},
    )
    mocker.patch.object(type(tracker_jobs), 'release_name', PropertyMock(return_value=Mock(
        fetch_info=Mock(),
    )))
    mocker.patch.object(type(tracker_jobs), 'release_name_job', PropertyMock(return_value=Mock(
        fetch_text=AsyncMock(),
    )))
    webdb = Mock()
    await tracker_jobs.update_release_name_from(webdb, 'tt123456')
    assert tracker_jobs.release_name_job.fetch_text.call_args_list == [
        call(
            coro=tracker_jobs.release_name.fetch_info.return_value,
            nonfatal_exceptions=(errors.RequestError,),
            default=str(tracker_jobs.release_name),
        ),
    ]
    assert tracker_jobs.release_name.fetch_info.call_args_list == [call(webdb=webdb, webdb_id='tt123456')]


def test_imdb_job(mocker):
    tracker_jobs = make_TestTrackerJobs(
        content_path='path/to/content',
        common_job_args={'home_directory': 'path/to/home', 'ignore_cache': 'mock bool'},
    )
    WebDbSearchJob_mock = mocker.patch('upsies.jobs.webdb.WebDbSearchJob')
    mocker.patch.object(tracker_jobs, 'make_precondition')
    imdb_job = tracker_jobs.imdb_job
    assert imdb_job is WebDbSearchJob_mock.return_value
    assert WebDbSearchJob_mock.call_args_list == [
        call(
            query='path/to/content',
            db=tracker_jobs.imdb,
            callbacks={
                'output': tracker_jobs._handle_imdb_id,
            },
            precondition=tracker_jobs.make_precondition.return_value,
            home_directory='path/to/home',
            ignore_cache='mock bool',
        ),
    ]
    assert tracker_jobs.make_precondition.call_args_list == [call('imdb_job')]

def test_imdb_job_is_singleton(mocker):
    mocker.patch('upsies.jobs.webdb.WebDbSearchJob', side_effect=(Mock(), Mock()))
    tracker_jobs = make_TestTrackerJobs(
        content_path='path/to/content',
        common_job_args={'home_directory': 'path/to/home', 'ignore_cache': 'mock bool'},
    )
    assert tracker_jobs.imdb_job is tracker_jobs.imdb_job


@pytest.mark.parametrize(
    argnames='imdb_id, imdb, poster_webdb, exp_start_poster_job_calls',
    argvalues=(
        ('imdb id', 'mock imdb', None, []),
        ('imdb id', 'mock imdb', 'mock imdb', [call._start_poster_job('mock imdb', 'imdb id')]),
        ('imdb id', 'mock imdb', 'mock otherdb', []),
    ),
)
def test_handle_imdb_id(imdb_id, imdb, poster_webdb, exp_start_poster_job_calls, mocker):
    tracker_jobs = make_TestTrackerJobs(
        content_path='path/to/content',
        common_job_args={'home_directory': 'path/to/home', 'ignore_cache': 'mock bool'},
    )

    mocks = Mock()
    mocks.attach_mock(mocker.patch.object(tracker_jobs.tracker, 'attach_task'), 'attach_task')
    mocks.attach_mock(mocker.patch.object(tracker_jobs, '_propagate_webdb_info', Mock()), '_propagate_webdb_info')
    mocks.attach_mock(mocker.patch.object(tracker_jobs, '_start_poster_job'), '_start_poster_job')
    mocker.patch.object(type(tracker_jobs), 'imdb', PropertyMock(return_value=imdb))
    mocker.patch.object(type(tracker_jobs), '_poster_webdb', PropertyMock(return_value=poster_webdb))

    tracker_jobs._handle_imdb_id(imdb_id)
    assert mocks.mock_calls == [
        call._propagate_webdb_info(tracker_jobs.imdb, imdb_id),
        call.attach_task(tracker_jobs._propagate_webdb_info.return_value),
    ] + exp_start_poster_job_calls


def test_imdb_id_property(mocker):
    tracker_jobs = make_TestTrackerJobs()
    mocker.patch.object(tracker_jobs, '_get_webdb_id_from_job')
    return_value = tracker_jobs.imdb_id
    assert return_value is tracker_jobs._get_webdb_id_from_job.return_value
    assert tracker_jobs._get_webdb_id_from_job.call_args_list == [call(tracker_jobs.imdb_job)]


def test_tmdb_job(mocker):
    WebDbSearchJob_mock = mocker.patch('upsies.jobs.webdb.WebDbSearchJob')
    webdb_mock = mocker.patch('upsies.utils.webdbs.webdb')
    tracker_jobs = make_TestTrackerJobs(
        content_path='path/to/content',
        common_job_args={'home_directory': 'path/to/home', 'ignore_cache': 'mock bool'},
    )
    mocker.patch.object(tracker_jobs, 'make_precondition')
    assert tracker_jobs.tmdb_job is WebDbSearchJob_mock.return_value
    assert WebDbSearchJob_mock.call_args_list == [
        call(
            query='path/to/content',
            db=webdb_mock.return_value,
            callbacks={
                'output': tracker_jobs._handle_tmdb_id,
            },
            precondition=tracker_jobs.make_precondition.return_value,
            home_directory='path/to/home',
            ignore_cache='mock bool',
        ),
    ]
    assert webdb_mock.call_args_list == [call('tmdb')]
    assert tracker_jobs.make_precondition.call_args_list == [call('tmdb_job')]


def test_tmdb_job_is_singleton(mocker):
    mocker.patch('upsies.jobs.webdb.WebDbSearchJob', side_effect=(Mock(), Mock()))
    tracker_jobs = make_TestTrackerJobs(
        content_path='path/to/content',
        common_job_args={'home_directory': 'path/to/home', 'ignore_cache': 'mock bool'},
    )
    assert tracker_jobs.tmdb_job is tracker_jobs.tmdb_job


@pytest.mark.parametrize(
    argnames='tmdb_id, tmdb, poster_webdb, exp_start_poster_job_calls',
    argvalues=(
        ('tmdb id', 'mock tmdb', None, []),
        ('tmdb id', 'mock tmdb', 'mock tmdb', [call._start_poster_job('mock tmdb', 'tmdb id')]),
        ('tmdb id', 'mock tmdb', 'mock otherdb', []),
    ),
)
def test_handle_tmdb_id(tmdb_id, tmdb, poster_webdb, exp_start_poster_job_calls, mocker):
    tracker_jobs = make_TestTrackerJobs(
        content_path='path/to/content',
        common_job_args={'home_directory': 'path/to/home', 'ignore_cache': 'mock bool'},
    )

    mocks = Mock()
    mocks.attach_mock(mocker.patch.object(tracker_jobs.tracker, 'attach_task'), 'attach_task')
    mocks.attach_mock(mocker.patch.object(tracker_jobs, '_propagate_webdb_info', Mock()), '_propagate_webdb_info')
    mocks.attach_mock(mocker.patch.object(tracker_jobs, '_start_poster_job'), '_start_poster_job')
    mocker.patch.object(type(tracker_jobs), 'tmdb', PropertyMock(return_value=tmdb))
    mocker.patch.object(type(tracker_jobs), '_poster_webdb', PropertyMock(return_value=poster_webdb))

    tracker_jobs._handle_tmdb_id(tmdb_id)
    assert mocks.mock_calls == [
        call._propagate_webdb_info(tracker_jobs.tmdb, tmdb_id),
        call.attach_task(tracker_jobs._propagate_webdb_info.return_value),
    ] + exp_start_poster_job_calls


def test_tmdb_id_property(mocker):
    tracker_jobs = make_TestTrackerJobs()
    mocker.patch.object(tracker_jobs, '_get_webdb_id_from_job')
    return_value = tracker_jobs.tmdb_id
    assert return_value is tracker_jobs._get_webdb_id_from_job.return_value
    assert tracker_jobs._get_webdb_id_from_job.call_args_list == [call(tracker_jobs.tmdb_job)]


def test_tvmaze_job(mocker):
    WebDbSearchJob_mock = mocker.patch('upsies.jobs.webdb.WebDbSearchJob')
    webdb_mock = mocker.patch('upsies.utils.webdbs.webdb')
    tracker_jobs = make_TestTrackerJobs(
        content_path='path/to/content',
        common_job_args={'home_directory': 'path/to/home', 'ignore_cache': 'mock bool'},
    )
    mocker.patch.object(tracker_jobs, 'make_precondition')
    assert tracker_jobs.tvmaze_job is WebDbSearchJob_mock.return_value
    assert WebDbSearchJob_mock.call_args_list == [
        call(
            query='path/to/content',
            db=webdb_mock.return_value,
            callbacks={
                'output': tracker_jobs._handle_tvmaze_id,
            },
            precondition=tracker_jobs.make_precondition.return_value,
            home_directory='path/to/home',
            ignore_cache='mock bool',
        ),
    ]
    assert webdb_mock.call_args_list == [call('tvmaze')]
    assert tracker_jobs.make_precondition.call_args_list == [call('tvmaze_job')]

def test_tvmaze_job_is_singleton(mocker):
    mocker.patch('upsies.jobs.webdb.WebDbSearchJob', side_effect=(Mock(), Mock()))
    tracker_jobs = make_TestTrackerJobs(
        content_path='path/to/content',
        common_job_args={'home_directory': 'path/to/home', 'ignore_cache': 'mock bool'},
    )
    assert tracker_jobs.tvmaze_job is tracker_jobs.tvmaze_job


@pytest.mark.parametrize(
    argnames='tvmaze_id, tvmaze, poster_webdb, exp_start_poster_job_calls',
    argvalues=(
        ('tvmaze id', 'mock tvmaze', None, []),
        ('tvmaze id', 'mock tvmaze', 'mock tvmaze', [call._start_poster_job('mock tvmaze', 'tvmaze id')]),
        ('tvmaze id', 'mock tvmaze', 'mock otherdb', []),
    ),
)
def test_handle_tvmaze_id(tvmaze_id, tvmaze, poster_webdb, exp_start_poster_job_calls, mocker):
    tracker_jobs = make_TestTrackerJobs(
        content_path='path/to/content',
        common_job_args={'home_directory': 'path/to/home', 'ignore_cache': 'mock bool'},
    )

    mocks = Mock()
    mocks.attach_mock(mocker.patch.object(tracker_jobs.tracker, 'attach_task'), 'attach_task')
    mocks.attach_mock(mocker.patch.object(tracker_jobs, '_propagate_webdb_info', Mock()), '_propagate_webdb_info')
    mocks.attach_mock(mocker.patch.object(tracker_jobs, '_start_poster_job'), '_start_poster_job')
    mocker.patch.object(type(tracker_jobs), 'tvmaze', PropertyMock(return_value=tvmaze))
    mocker.patch.object(type(tracker_jobs), '_poster_webdb', PropertyMock(return_value=poster_webdb))

    tracker_jobs._handle_tvmaze_id(tvmaze_id)
    assert mocks.mock_calls == [
        call._propagate_webdb_info(tracker_jobs.tvmaze, tvmaze_id),
        call.attach_task(tracker_jobs._propagate_webdb_info.return_value),
    ] + exp_start_poster_job_calls


def test_tvmaze_id_property(mocker):
    tracker_jobs = make_TestTrackerJobs()
    mocker.patch.object(tracker_jobs, '_get_webdb_id_from_job')
    return_value = tracker_jobs.tvmaze_id
    assert return_value is tracker_jobs._get_webdb_id_from_job.return_value
    assert tracker_jobs._get_webdb_id_from_job.call_args_list == [call(tracker_jobs.tvmaze_job)]


@pytest.mark.parametrize(
    argnames='webdb_job, exp_return_value',
    argvalues=(
        (Mock(is_finished=False, output=[]), None),
        (Mock(is_finished=False, output=['mock id']), None),
        (Mock(is_finished=False, output=['mock id', 'another mock id']), None),
        (Mock(is_finished=True, output=[]), None),
        (Mock(is_finished=True, output=['mock id']), 'mock id'),
        (Mock(is_finished=True, output=['mock id', 'another mock id']), 'mock id'),
    ),
)
def test_get_webdb_id_from_job(webdb_job, exp_return_value, mocker):
    tracker_jobs = make_TestTrackerJobs()
    return_value = tracker_jobs._get_webdb_id_from_job(webdb_job)
    assert return_value == exp_return_value


@pytest.mark.parametrize('english_title', ('', 'English Title'))
@pytest.mark.parametrize('origin_job_is_finished', (False, True))
@pytest.mark.asyncio
async def test_propagate_webdb_info(origin_job_is_finished, english_title, mocker):
    tracker_jobs = make_TestTrackerJobs(
        content_path='path/to/content',
        common_job_args={'home_directory': 'path/to/home', 'ignore_cache': 'mock bool'},
    )

    mocks = Mock(
        update_release_name_from=AsyncMock(),
    )
    mocker.patch.object(tracker_jobs, 'update_release_name_from', mocks.update_release_name_from)
    mocker.patch('upsies.utils.webdbs.webdb_names', mocks.webdb_names)
    mocks.webdb_names.return_value = ('adb', 'bdb', 'cdb', 'ddb')

    mocker.patch('upsies.utils.webdbs.Query', mocks.Query)

    def make_db(name, type, title_english, title_original, year):
        return types.SimpleNamespace(
            name=name,
            type=AsyncMock(return_value=type),
            title_english=AsyncMock(return_value=title_english),
            title_original=AsyncMock(return_value=title_original),
            year=AsyncMock(return_value=year),
        )

    adb = make_db('adb', 'type from adb', english_title, 'original title from adb', 'year from adb')
    bdb = make_db('bdb', 'type from bdb', 'english title from bdb', 'original title from bdb', 'year from bdb')
    cdb = make_db('cdb', 'type from cdb', 'english title from cdb', 'original title from cdb', 'year from cdb')
    ddb = make_db('ddb', 'type from ddb', 'english title from ddb', 'original title from ddb', 'year from ddb')
    mocker.patch.object(type(tracker_jobs), 'adb', PropertyMock(return_value=adb), create=True)
    mocker.patch.object(type(tracker_jobs), 'bdb', PropertyMock(return_value=bdb), create=True)
    mocker.patch.object(type(tracker_jobs), 'cdb', PropertyMock(return_value=cdb), create=True)
    mocker.patch.object(type(tracker_jobs), 'ddb', PropertyMock(return_value=ddb), create=True)

    def make_db_job(name, is_enabled, is_finished, search):
        return types.SimpleNamespace(
            name=name,
            is_enabled=is_enabled,
            is_finished=is_finished,
            search=search,
            query=getattr(mocks, name[:name.index('-')]).query,
        )

    adb_job = make_db_job('adb-id', True, origin_job_is_finished, mocks.adb.search)
    bdb_job = make_db_job('bdb-id', False, True, mocks.bdb.search)
    cdb_job = make_db_job('cdb-id', True, False, mocks.cdb.search)
    ddb_job = make_db_job('ddb-id', True, True, mocks.ddb.search)
    mocker.patch.object(type(tracker_jobs), 'adb_job', PropertyMock(return_value=adb_job), create=True)
    mocker.patch.object(type(tracker_jobs), 'bdb_job', PropertyMock(return_value=bdb_job), create=True)
    mocker.patch.object(type(tracker_jobs), 'cdb_job', PropertyMock(return_value=cdb_job), create=True)
    mocker.patch.object(type(tracker_jobs), 'ddb_job', PropertyMock(return_value=ddb_job), create=True)

    await tracker_jobs._propagate_webdb_info(tracker_jobs.adb, '123456')

    exp_title = english_title or 'original title from adb'
    assert mocks.mock_calls == [
        call.webdb_names(),
        call.Query(type='type from adb', title=exp_title, year='year from adb'),
        call.cdb.query.update(mocks.Query.return_value),
        call.update_release_name_from(tracker_jobs.adb, '123456'),
    ]


def test_screenshots_job(mocker):
    ScreenshotsJob_mock = mocker.patch('upsies.jobs.screenshots.ScreenshotsJob')
    tracker_jobs = make_TestTrackerJobs(
        content_path='path/to/content',
        common_job_args={'home_directory': 'path/to/home', 'ignore_cache': 'mock bool'},
    )
    mocker.patch.object(tracker_jobs, 'make_precondition')
    mocker.patch.object(type(tracker_jobs), 'screenshots_count', PropertyMock())
    mocker.patch.object(type(tracker_jobs), 'screenshots_from_all_videos', PropertyMock())
    mocker.patch.object(type(tracker_jobs), 'create_torrent_job', PropertyMock())
    assert tracker_jobs.screenshots_job is ScreenshotsJob_mock.return_value
    assert ScreenshotsJob_mock.call_args_list == [
        call(
            content_path='path/to/content',
            exclude_files=tracker_jobs.exclude_files,
            count=tracker_jobs.screenshots_count,
            from_all_videos=tracker_jobs.screenshots_from_all_videos,
            optimize=tracker_jobs._screenshots_optimization,
            precondition=tracker_jobs.make_precondition.return_value,
            home_directory='path/to/home',
            ignore_cache='mock bool',
        ),
    ]
    assert tracker_jobs.make_precondition.call_args_list == [call('screenshots_job')]

def test_screenshots_job_is_singleton(mocker):
    mocker.patch('upsies.jobs.screenshots.ScreenshotsJob', side_effect=(Mock(), Mock()))
    tracker_jobs = make_TestTrackerJobs(
        content_path='path/to/content',
        common_job_args={'home_directory': 'path/to/home', 'ignore_cache': 'mock bool'},
    )
    mocker.patch.object(type(tracker_jobs), 'create_torrent_job', PropertyMock())
    assert tracker_jobs.screenshots_job is tracker_jobs.screenshots_job


def test_screenshots_count(mocker):
    tracker_jobs = make_TestTrackerJobs(
        content_path='path/to/content',
        common_job_args={'home_directory': 'path/to/home', 'ignore_cache': 'mock bool'},
        options={'screenshots': 123},
    )
    assert tracker_jobs.screenshots_count == 123


def test_screenshots_from_all_videos(mocker):
    tracker_jobs = make_TestTrackerJobs(
        content_path='path/to/content',
        common_job_args={'home_directory': 'path/to/home', 'ignore_cache': 'mock bool'},
    )
    assert tracker_jobs.screenshots_from_all_videos is False


def test_image_host_config():
    assert TrackerJobsBase.image_host_config == {}


@pytest.mark.parametrize(
    argnames='image_host, screenshots_job, exp_upload_screenshots_job_is_None',
    argvalues=(
        (Mock(), Mock(), False),
        (None, Mock(), True),
        (Mock(), None, True),
    ),
)
def test_upload_screenshots_job(image_host, screenshots_job, exp_upload_screenshots_job_is_None, mocker):
    ImageHostJob_mock = mocker.patch('upsies.jobs.imghost.ImageHostJob')
    mocker.patch('upsies.jobs.screenshots.ScreenshotsJob')
    tracker_jobs = make_TestTrackerJobs(
        content_path='path/to/content',
        common_job_args={'home_directory': 'path/to/home', 'ignore_cache': 'mock bool'},
        image_host=image_host,
    )
    mocker.patch.object(type(tracker_jobs), 'screenshots_job', PropertyMock(return_value=screenshots_job))
    mocker.patch.object(tracker_jobs, 'make_precondition')
    if exp_upload_screenshots_job_is_None:
        assert tracker_jobs.upload_screenshots_job is None
        assert ImageHostJob_mock.call_args_list == []
        assert tracker_jobs.make_precondition.call_args_list == []
    else:
        assert tracker_jobs.upload_screenshots_job is ImageHostJob_mock.return_value
        assert ImageHostJob_mock.call_args_list == [call(
            imghost=image_host,
            precondition=tracker_jobs.make_precondition.return_value,
            home_directory='path/to/home',
            ignore_cache='mock bool',
        )]
        # ScreenshotsJob also registers a callback for "timestamps"
        assert tracker_jobs.screenshots_job.signal.register.call_args_list == [
            call('screenshots_total', tracker_jobs.upload_screenshots_job.set_images_total),
            call('output', tracker_jobs.upload_screenshots_job.enqueue),
            call('finished', tracker_jobs.finalize_upload_screenshots_job),
        ]
        assert tracker_jobs.make_precondition.call_args_list == [call('upload_screenshots_job')]

def test_upload_screenshots_job_is_singleton(mocker):
    mocker.patch('upsies.jobs.imghost.ImageHostJob', side_effect=(Mock(), Mock()))
    mocker.patch('upsies.jobs.screenshots.ScreenshotsJob')
    tracker_jobs = make_TestTrackerJobs(
        content_path='path/to/content',
        common_job_args={'home_directory': 'path/to/home', 'ignore_cache': 'mock bool'},
        image_host=Mock(),
    )
    assert tracker_jobs.upload_screenshots_job is tracker_jobs.upload_screenshots_job


def test_poster_job(mocker):
    PosterJob_mock = mocker.patch('upsies.jobs.poster.PosterJob')
    tracker_jobs = make_TestTrackerJobs(
        content_path='path/to/content',
        common_job_args={'home_directory': 'path/to/home', 'ignore_cache': 'mock bool'},
    )
    mocker.patch.object(tracker_jobs, 'make_precondition')
    assert tracker_jobs.poster_job is PosterJob_mock.return_value
    assert PosterJob_mock.call_args_list == [
        call(
            autostart=False,
            webdb_id=None,
            webdb=None,
            imghost=tracker_jobs.image_host,
            width=tracker_jobs.poster_max_width,
            height=tracker_jobs.poster_max_height,
            precondition=tracker_jobs.make_precondition.return_value,
            home_directory='path/to/home',
            ignore_cache='mock bool',
        ),
    ]
    assert tracker_jobs.make_precondition.call_args_list == [call('poster_job')]

def test_poster_job_is_singleton(mocker):
    mocker.patch('upsies.jobs.poster.PosterJob', side_effect=(Mock(), Mock()))
    tracker_jobs = make_TestTrackerJobs(
        content_path='path/to/content',
        common_job_args={'home_directory': 'path/to/home', 'ignore_cache': 'mock bool'},
    )
    assert tracker_jobs.poster_job is tracker_jobs.poster_job


def test_start_poster_job(mocker):
    tracker_jobs = make_TestTrackerJobs(
        content_path='path/to/content',
        common_job_args={'home_directory': 'path/to/home', 'ignore_cache': 'mock bool'},
    )

    mocker.patch.object(tracker_jobs.poster_job, 'start')
    mocker.patch.object(type(tracker_jobs.release_name), 'only_season', PropertyMock(return_value='3'))

    tracker_jobs._start_poster_job('mock webdb', 'webdb id')
    assert tracker_jobs.poster_job.start.call_args_list == [call(
        webdb='mock webdb',
        webdb_id='webdb id',
        season='3',
    )]


@pytest.mark.parametrize(
    argnames='is_enabled, is_started, in_jobs_before_upload',
    argvalues=(
        (False, False, False),

        (False, False, True),
        (False, True, False),
        (True, False, False),

        (False, True, True),
        # (True, False, True),  # poster_job is enabled
        (True, True, False),

        (True, True, True),
    ),
)
def test_poster_webdb_when_poster_job_is_disabled(is_enabled, is_started, in_jobs_before_upload, mocker):
    tracker_jobs = make_TestTrackerJobs(
        content_path='path/to/content',
        common_job_args={'home_directory': 'path/to/home', 'ignore_cache': 'mock bool'},
    )

    mocker.patch.object(type(tracker_jobs), 'poster_job', PropertyMock(return_value=Mock(
        is_enabled=is_enabled,
        is_started=is_started,
    )))
    if in_jobs_before_upload:
        jobs_before_upload = ('foo', tracker_jobs.poster_job, 'bar')
    else:
        jobs_before_upload = ('foo', 'bar')
    mocker.patch.object(type(tracker_jobs), 'jobs_before_upload', PropertyMock(return_value=jobs_before_upload))

    assert tracker_jobs._poster_webdb is None

@pytest.mark.parametrize(
    argnames=(
        'tvmaze_is_enabled, tvmaze_in_jobs_before_upload, release_type, '
        'imdb_is_enabled, imdb_in_jobs_before_upload, '
        'tmdb_is_enabled, tmdb_in_jobs_before_upload, '
        'exp_result'
    ),
    argvalues=(
        pytest.param(
            True, True, utils.types.ReleaseType.season,  # TVmaze
            True, True,  # IMDb
            True, True,  # TMDb
            'tvmaze',
            id='tvmaze for season',
        ),
        pytest.param(
            True, True, utils.types.ReleaseType.episode,  # TVmaze
            True, True,  # IMDb
            True, True,  # TMDb
            'tvmaze',
            id='tvmaze for episode',
        ),
        pytest.param(
            True, True, utils.types.ReleaseType.movie,  # TVmaze
            True, True,  # IMDb
            True, True,  # TMDb
            'imdb',
            id='imdb for movie',
        ),
        pytest.param(
            True, False, utils.types.ReleaseType.series,  # TVmaze
            True, True,  # IMDb
            True, True,  # TMDb
            'imdb',
            id='imdb if tvmaze is not used',
        ),
        pytest.param(
            False, True, utils.types.ReleaseType.series,  # TVmaze
            True, True,  # IMDb
            True, True,  # TMDb
            'imdb',
            id='imdb if tvmaze is not enabled',
        ),
        pytest.param(
            False, False, utils.types.ReleaseType.movie,  # TVmaze
            True, False,  # IMDb
            True, True,  # TMDb
            'tmdb',
            id='tmdb if imdb is not used',
        ),
        pytest.param(
            False, False, utils.types.ReleaseType.movie,  # TVmaze
            False, True,  # IMDb
            True, True,  # TMDb
            'tmdb',
            id='tmdb if imdb is not enabled',
        ),
        pytest.param(
            True, False, utils.types.ReleaseType.movie,  # TVmaze
            True, False,  # IMDb
            True, False,  # TMDb
            RuntimeError('You must add a WebDbSearchJob instance to jobs_before_upload'),
            id='error if no webdb job is used',
        ),
        pytest.param(
            False, True, utils.types.ReleaseType.movie,  # TVmaze
            False, True,  # IMDb
            False, True,  # TMDb
            RuntimeError('You must add a WebDbSearchJob instance to jobs_before_upload'),
            id='error if no webdb job is enabled',
        ),
    ),
)
def test_poster_webdb_returns(
        tvmaze_is_enabled, tvmaze_in_jobs_before_upload, release_type,
        imdb_is_enabled, imdb_in_jobs_before_upload,
        tmdb_is_enabled, tmdb_in_jobs_before_upload,
        exp_result,
        mocker,
):
    tracker_jobs = make_TestTrackerJobs(
        content_path='path/to/content',
        common_job_args={'home_directory': 'path/to/home', 'ignore_cache': 'mock bool'},
    )

    mocker.patch.object(type(tracker_jobs), 'poster_job', PropertyMock(return_value=Mock(
        is_enabled=True,
        is_started=False,
    )))
    mocker.patch.object(type(tracker_jobs), 'imdb_job', PropertyMock(return_value=Mock(is_enabled=imdb_is_enabled)))
    mocker.patch.object(type(tracker_jobs), 'tmdb_job', PropertyMock(return_value=Mock(is_enabled=tmdb_is_enabled)))
    mocker.patch.object(type(tracker_jobs), 'tvmaze_job', PropertyMock(return_value=Mock(is_enabled=tvmaze_is_enabled)))
    mocker.patch.object(type(tracker_jobs.release_name), 'type', PropertyMock(return_value=release_type))

    jobs_before_upload = ['foo', tracker_jobs.poster_job, 'bar']
    if imdb_in_jobs_before_upload:
        jobs_before_upload.insert(1, tracker_jobs.imdb_job)
    if tmdb_in_jobs_before_upload:
        jobs_before_upload.insert(1, tracker_jobs.tmdb_job)
    if tvmaze_in_jobs_before_upload:
        jobs_before_upload.insert(1, tracker_jobs.tvmaze_job)
    print(jobs_before_upload)
    mocker.patch.object(type(tracker_jobs), 'jobs_before_upload', PropertyMock(return_value=jobs_before_upload))

    if isinstance(exp_result, Exception):
        with pytest.raises(type(exp_result), match=rf'^{re.escape(str(exp_result))}$'):
            tracker_jobs._poster_webdb
    else:
        assert tracker_jobs._poster_webdb is getattr(tracker_jobs, exp_result)


def test_finalize_upload_screenshots_job(mocker):
    mocker.patch('upsies.jobs.screenshots.ScreenshotsJob', side_effect=(Mock(), Mock()))
    tracker_jobs = make_TestTrackerJobs(
        content_path='path/to/content',
        tracker='mock tracker',
        common_job_args={'home_directory': 'path/to/home', 'ignore_cache': 'mock bool'},
        btclient='bittorrent client mock',
    )
    mocker.patch.object(tracker_jobs, 'upload_screenshots_job')
    tracker_jobs.finalize_upload_screenshots_job('mock job')
    assert tracker_jobs.upload_screenshots_job.finalize.call_args_list == [call()]


def test_mediainfo_job(mocker):
    MediainfoJob_mock = mocker.patch('upsies.jobs.mediainfo.MediainfoJob')
    tracker_jobs = make_TestTrackerJobs(
        content_path='path/to/content',
        common_job_args={'home_directory': 'path/to/home', 'ignore_cache': 'mock bool'},
    )
    mocker.patch.object(type(tracker_jobs), 'mediainfo_from_all_videos', PropertyMock(
        return_value='maybe from_all_videos?',
    )),
    mocker.patch.object(type(tracker_jobs), 'exclude_files', PropertyMock(
        return_value='exclude these files!',
    )),
    mocker.patch.object(tracker_jobs, 'make_precondition')
    assert tracker_jobs.mediainfo_job is MediainfoJob_mock.return_value
    assert MediainfoJob_mock.call_args_list == [
        call(
            content_path='path/to/content',
            from_all_videos='maybe from_all_videos?',
            exclude_files='exclude these files!',
            precondition=tracker_jobs.make_precondition.return_value,
            home_directory='path/to/home',
            ignore_cache='mock bool',
        ),
    ]
    assert tracker_jobs.make_precondition.call_args_list == [call('mediainfo_job')]

def test_mediainfo_job_is_singleton(mocker):
    mocker.patch('upsies.jobs.mediainfo.MediainfoJob', side_effect=(Mock(), Mock()))
    tracker_jobs = make_TestTrackerJobs(
        content_path='path/to/content',
        common_job_args={'home_directory': 'path/to/home', 'ignore_cache': 'mock bool'},
    )
    assert tracker_jobs.mediainfo_job is tracker_jobs.mediainfo_job


def test_mediainfos_and_screenshots(mocker):
    tracker_jobs = make_TestTrackerJobs(
        content_path='path/to/content',
        common_job_args={'home_directory': 'path/to/home', 'ignore_cache': 'mock bool'},
    )

    mocker.patch.object(type(tracker_jobs), 'mediainfo_job', PropertyMock(return_value=Mock(mediainfos_by_file={
        'path/to/foo.mkv': '<mediainfo for foo.mkv>',
        'path/to/bar/VIDEO_TS/VTS_01_0.IFO': '<mediainfo for bar/VTS_01_0.IFO>',
        'path/to/bar/VIDEO_TS/VTS_01_1.VOB': '<mediainfo for bar/VTS_01_1.VOB>',
    })))
    mocker.patch.object(type(tracker_jobs), 'screenshots_job', PropertyMock(return_value=Mock(screenshots_by_file={
        'path/to/foo.mkv': ('path/to/foo.1.png', 'path/to/foo.2.png'),
        'path/to/bar/VIDEO_TS/VTS_01_1.VOB': ('VTS_01_1.VOB.1.jpg',),
        'path/to/baz.mkv': ('path/to/baz.1.png', 'path/to/baz.2.png', 'path/to/baz.3.png'),
    })))
    mocker.patch.object(type(tracker_jobs), 'upload_screenshots_job', PropertyMock(return_value=Mock(urls_by_file={
        'path/to/foo.1.png': 'https://foo.1.png',
        'path/to/foo.2.png': 'https://foo.2.png',
        'VTS_01_1.VOB.1.jpg': 'https://VTS_01_1.VOB.1.jpg',
        'path/to/baz.1.png': 'https://baz.1.png',
        'path/to/baz.2.png': 'https://baz.2.png',
    })))

    assert tracker_jobs.mediainfos_and_screenshots == {
        'path/to/foo.mkv': {
            'mediainfo': '<mediainfo for foo.mkv>',
            'screenshot_urls': ['https://foo.1.png', 'https://foo.2.png'],
        },
        'path/to/bar/VIDEO_TS/VTS_01_0.IFO': {
            'mediainfo': '<mediainfo for bar/VTS_01_0.IFO>',
            'screenshot_urls': [],
        },
        'path/to/bar/VIDEO_TS/VTS_01_1.VOB': {
            'mediainfo': '<mediainfo for bar/VTS_01_1.VOB>',
            'screenshot_urls': ['https://VTS_01_1.VOB.1.jpg'],
        },
        'path/to/baz.mkv': {
            'mediainfo': '',
            'screenshot_urls': ['https://baz.1.png', 'https://baz.2.png'],
        },
    }


@pytest.mark.parametrize(
    argnames='options, common_job_args, exp_kwargs',
    argvalues=(
        ({}, {'ignore_cache': True}, {'ignore_cache': True, 'force': None}),
        ({}, {'ignore_cache': False}, {'ignore_cache': True, 'force': None}),
        ({'is_scene': None}, {'ignore_cache': True}, {'ignore_cache': True, 'force': None}),
        ({'is_scene': None}, {'ignore_cache': False}, {'ignore_cache': True, 'force': None}),
        ({'is_scene': True}, {'ignore_cache': True}, {'ignore_cache': True, 'force': True}),
        ({'is_scene': True}, {'ignore_cache': False}, {'ignore_cache': True, 'force': True}),
        ({'is_scene': False}, {'ignore_cache': True}, {'ignore_cache': True, 'force': False}),
        ({'is_scene': False}, {'ignore_cache': False}, {'ignore_cache': True, 'force': False}),
    ),
    ids=lambda v: repr(v),
)
def test_scene_check_job(options, common_job_args, exp_kwargs, mocker):
    SceneCheckJob_mock = mocker.patch('upsies.jobs.scene.SceneCheckJob')
    tracker_jobs = make_TestTrackerJobs(
        content_path='path/to/content',
        common_job_args=common_job_args,
    )
    mocker.patch.object(type(tracker_jobs), 'options', PropertyMock(return_value=options))
    mocker.patch.object(tracker_jobs, 'make_precondition')
    assert tracker_jobs.scene_check_job is SceneCheckJob_mock.return_value
    assert SceneCheckJob_mock.call_args_list == [call(
        content_path='path/to/content',
        precondition=tracker_jobs.make_precondition.return_value,
        **exp_kwargs,
    )]
    assert tracker_jobs.make_precondition.call_args_list == [call('scene_check_job')]

def test_scene_check_job_is_singleton(mocker):
    mocker.patch('upsies.jobs.scene.SceneCheckJob', side_effect=(Mock(), Mock()))
    tracker_jobs = make_TestTrackerJobs(
        content_path='path/to/content',
        common_job_args={'home_directory': 'path/to/home', 'ignore_cache': 'mock bool'},
    )
    assert tracker_jobs.scene_check_job is tracker_jobs.scene_check_job


@pytest.mark.parametrize(
    argnames='job_attr, jobs_before_upload, jobs_after_upload, isolated_jobs, custom_precondition, exp_return_value',
    argvalues=(
        ('foo_job', (), (), (), None, False),
        ('foo_job', (), (), (), Mock(return_value=True), False),
        ('foo_job', (), (), (), Mock(return_value=False), False),

        ('foo_job', ('foo_job', 'bar_job'), ('baz_job',), (), None, True),
        ('foo_job', ('foo_job', 'bar_job'), ('baz_job',), (), Mock(return_value=True), True),
        ('foo_job', ('foo_job', 'bar_job'), ('baz_job',), (), Mock(return_value=False), False),

        ('bar_job', ('foo_job', 'bar_job'), ('baz_job',), (), None, True),
        ('bar_job', ('foo_job', 'bar_job'), ('baz_job',), (), Mock(return_value=True), True),
        ('bar_job', ('foo_job', 'bar_job'), ('baz_job',), (), Mock(return_value=False), False),

        ('baz_job', ('foo_job', 'bar_job'), ('baz_job',), (), None, True),
        ('baz_job', ('foo_job', 'bar_job'), ('baz_job',), (), Mock(return_value=True), True),
        ('baz_job', ('foo_job', 'bar_job'), ('baz_job',), (), Mock(return_value=False), False),

        ('baz_job', ('foo_job',), ('bar_job',), (), None, False),
        ('baz_job', ('foo_job',), ('bar_job',), (), Mock(return_value=True), False),
        ('baz_job', ('foo_job',), ('bar_job',), (), Mock(return_value=False), False),

        ('bar_job', (), ('bar_job',), (), None, True),
        ('bar_job', (), ('bar_job',), (), Mock(return_value=True), True),
        ('bar_job', (), ('bar_job',), (), Mock(return_value=False), False),

        ('foo_job', ('foo_job', 'bar_job'), ('baz_job',), ('foo_job',), None, True),
        ('foo_job', ('foo_job', 'bar_job'), ('baz_job',), ('foo_job',), Mock(return_value=True), True),
        ('foo_job', ('foo_job', 'bar_job'), ('baz_job',), ('foo_job',), Mock(return_value=False), False),

        ('bar_job', ('foo_job', 'bar_job'), ('baz_job',), ('foo_job',), None, False),
        ('bar_job', ('foo_job', 'bar_job'), ('baz_job',), ('foo_job',), Mock(return_value=True), False),
        ('bar_job', ('foo_job', 'bar_job'), ('baz_job',), ('foo_job',), Mock(return_value=False), False),

        ('bar_job', ('foo_job', 'bar_job'), ('baz_job',), ('foo_job', 'bar_job'), None, True),
        ('bar_job', ('foo_job', 'bar_job'), ('baz_job',), ('foo_job', 'bar_job'), Mock(return_value=True), True),
        ('bar_job', ('foo_job', 'bar_job'), ('baz_job',), ('foo_job', 'bar_job'), Mock(return_value=False), False),

        ('baz_job', ('foo_job', 'bar_job'), ('baz_job',), ('foo_job', 'bar_job'), None, False),
        ('baz_job', ('foo_job', 'bar_job'), ('baz_job',), ('foo_job', 'bar_job'), Mock(return_value=True), False),
        ('baz_job', ('foo_job', 'bar_job'), ('baz_job',), ('foo_job', 'bar_job'), Mock(return_value=False), False),
    ),
    ids=lambda v: str(v),
)
def test_make_precondition(job_attr, jobs_before_upload, jobs_after_upload, isolated_jobs, custom_precondition,
                           exp_return_value, mocker):
    tracker_jobs = make_TestTrackerJobs()
    for attr in ('foo_job', 'bar_job', 'baz_job'):
        mocker.patch.object(type(tracker_jobs), attr, PropertyMock(), create=True)
    mocker.patch.object(type(tracker_jobs), 'jobs_before_upload', PropertyMock(
        return_value=[getattr(tracker_jobs, attr) for attr in jobs_before_upload],
    ))
    mocker.patch.object(type(tracker_jobs), 'jobs_after_upload', PropertyMock(
        return_value=[getattr(tracker_jobs, attr) for attr in jobs_after_upload],
    ))
    mocker.patch.object(type(tracker_jobs), 'isolated_jobs', PropertyMock(
        return_value=[getattr(tracker_jobs, attr) for attr in isolated_jobs],
    ))
    precondition = tracker_jobs.make_precondition(job_attr, precondition=custom_precondition)
    return_value = precondition()
    assert return_value == exp_return_value


@pytest.mark.parametrize(
    argnames='job, slice, default, exp_result',
    argvalues=(
        (
            Mock(is_finished=True, output=('foo', 'bar', 'baz')),
            None,
            TrackerJobsBase._NO_DEFAULT,
            ('foo', 'bar', 'baz'),
        ),
        (
            Mock(is_finished=True, output=('foo', 'bar', 'baz')),
            0,
            'ignored default',
            'foo',
        ),
        (
            Mock(is_finished=True, output=('foo', 'bar', 'baz')),
            1,
            TrackerJobsBase._NO_DEFAULT,
            'bar',
        ),
        (
            Mock(is_finished=True, output=('foo', 'bar', 'baz')),
            2,
            'ignored default',
            'baz',
        ),
        (
            Mock(is_finished=True, output=('foo', 'bar', 'baz')),
            builtins.slice(1, 3),
            TrackerJobsBase._NO_DEFAULT,
            ('bar', 'baz'),
        ),

        # Job is not Finished
        (
            Mock(is_finished=False, output=('foo', 'bar', 'baz')),
            None,
            TrackerJobsBase._NO_DEFAULT,
            RuntimeError('Unfinished job: asdf'),
        ),
        (
            Mock(is_finished=False, output=('foo', 'bar', 'baz')),
            None,
            'my default',
            'my default',
        ),
        (
            Mock(is_finished=False, output=('foo', 'bar', 'baz')),
            None,
            None,
            None,
        ),

        # Slicing output fails with IndexError
        (
            Mock(is_finished=True, output=('foo', 'bar', 'baz')),
            10,
            TrackerJobsBase._NO_DEFAULT,
            RuntimeError("Job finished with insufficient output: asdf: ('foo', 'bar', 'baz')"),
        ),
        (
            Mock(is_finished=True, output=('foo', 'bar', 'baz')),
            10,
            'my default',
            RuntimeError("Job finished with insufficient output: asdf: ('foo', 'bar', 'baz')"),
        ),
        (
            Mock(is_finished=True, output=('foo', 'bar', 'baz')),
            10,
            None,
            RuntimeError("Job finished with insufficient output: asdf: ('foo', 'bar', 'baz')"),
        ),
    ),
    ids=lambda v: repr(v),
)
@pytest.mark.asyncio
async def test_get_job_output(job, slice, default, exp_result, mocker):
    job.configure_mock(name='asdf')
    tracker_jobs = make_TestTrackerJobs(
        content_path='path/to/content',
        common_job_args={'home_directory': 'path/to/home', 'ignore_cache': 'mock bool'},
    )
    if isinstance(exp_result, Exception):
        with pytest.raises(type(exp_result), match=rf'^{re.escape(str(exp_result))}$'):
            tracker_jobs.get_job_output(job, slice, default=default)
    else:
        output = tracker_jobs.get_job_output(job, slice, default=default)
        assert output == exp_result


@pytest.mark.parametrize(
    argnames='job, attribute, default, exp_result',
    argvalues=(
        (
            Mock(is_finished=True, foo='bar'),
            'foo',
            TrackerJobsBase._NO_DEFAULT,
            'bar',
        ),
        (
            Mock(is_finished=True, foo='bar'),
            'foo',
            'ignored default',
            'bar',
        ),
        (
            Mock(is_finished=False, foo='bar'),
            'foo',
            TrackerJobsBase._NO_DEFAULT,
            RuntimeError('Unfinished job: asdf'),
        ),
        (
            Mock(is_finished=False, foo='bar'),
            'foo',
            'my default',
            'my default',
        ),
        (
            Mock(is_finished=True, foo='bar', spec=()),
            'no_such_attribute',
            TrackerJobsBase._NO_DEFAULT,
            AttributeError('no_such_attribute'),
        ),
    ),
    ids=lambda v: repr(v),
)
@pytest.mark.asyncio
async def test_get_job_attribute(job, attribute, default, exp_result, mocker):
    job.configure_mock(name='asdf')
    tracker_jobs = make_TestTrackerJobs(
        content_path='path/to/content',
        common_job_args={'home_directory': 'path/to/home', 'ignore_cache': 'mock bool'},
    )
    if isinstance(exp_result, Exception):
        with pytest.raises(type(exp_result), match=rf'{re.escape(str(exp_result))}'):
            tracker_jobs.get_job_attribute(job, attribute, default=default)
    else:
        value = tracker_jobs.get_job_attribute(job, attribute, default=default)
        assert value == exp_result


@pytest.mark.parametrize(
    argnames='file_path, content_path, exp_result',
    argvalues=(
        (
            '/path/to/somewhere',
            '/this/is/completely/different',
            ValueError(r"'[\w+/]*to/somewhere' is not a subpath of '[\w+/]*completely'"),
        ),
        (
            '{tmp_path}/file.mkv',
            '{tmp_path}/file.mkv',
            'file.mkv',
        ),
        (
            '{tmp_path}/content/file_in_directory.mkv',
            '{tmp_path}/content',
            'content/file_in_directory.mkv',
        ),
        (
            '{tmp_path}/content/extras/file_in_subdirectory.mkv',
            '{tmp_path}/content',
            'content/extras/file_in_subdirectory.mkv',
        ),
    ),
    ids=lambda v: repr(v),
)
@pytest.mark.parametrize('file_path_is_relative', (False, True), ids=('file_path_is_absolute', 'file_path_is_relative'))
@pytest.mark.parametrize('content_path_is_relative', (False, True), ids=('content_path_is_absolute', 'content_path_is_relative'))
def test_get_relative_path(file_path, content_path, exp_result,
                           file_path_is_relative, content_path_is_relative,
                           tmp_path, mocker):

    def write(path):
        path = pathlib.Path(path.format(tmp_path=str(tmp_path)))
        try:
            path.relative_to(tmp_path)
        except ValueError:
            print(':::  not creating:', path)
        else:
            path.parent.mkdir(parents=True, exist_ok=True)
            if path.suffix:
                print(':::  creating file:', path)
                path.write_bytes(b'content data')
            else:
                print(':::  creating directory:', path)
                path.mkdir(exist_ok=True)
        return path

    content_path = write(content_path)
    file_path = write(file_path)

    cwd = None
    orig_cwd = pathlib.Path(os.getcwd())
    if content_path_is_relative:
        cwd = content_path.parent.parent
        content_path = content_path.relative_to(cwd)
    if file_path_is_relative:
        if not cwd:
            cwd = file_path.parent.parent
        try:
            file_path = file_path.relative_to(cwd)
        except ValueError:
            pass

    print('___ content_path:', content_path)
    print('___ file_path:', file_path)
    print('___ cwd:', cwd)
    if cwd:
        if cwd.exists():
            os.chdir(cwd)
    try:
        tracker_jobs = make_TestTrackerJobs(
            content_path=str(content_path),
            common_job_args={'home_directory': 'path/to/home', 'ignore_cache': 'mock bool'},
        )
        if isinstance(exp_result, Exception):
            exp_msg = type(exp_result)(str(exp_result).format(tmp_path=str(tmp_path)))
            with pytest.raises(type(exp_result), match=rf'^{exp_msg}$'):
                tracker_jobs.get_relative_file_path(str(file_path))
        else:
            relative_path = tracker_jobs.get_relative_file_path(str(file_path))
            assert relative_path == exp_result
    finally:
        if orig_cwd.exists():
            os.chdir(orig_cwd)


class ImageUrl(str):
    def __new__(cls, *args, thumbnail=True, **kwargs):
        self = super().__new__(cls, *args, **kwargs)
        self._thumbnail = thumbnail
        return self

    @property
    def thumbnail_url(self):
        if self._thumbnail:
            return f'thumb_{self}'

@pytest.mark.parametrize(
    argnames='screenshots, columns, horizontal_spacer, vertical_spacer, exp_bbcode',
    argvalues=(
        (
            (ImageUrl('a.png'), ImageUrl('b.png'), ImageUrl('c.png')),
            1, ' | ', '\n-\n',
            (
                '[url=a.png][img]thumb_a.png[/img][/url]'
                '\n-\n'
                '[url=b.png][img]thumb_b.png[/img][/url]'
                '\n-\n'
                '[url=c.png][img]thumb_c.png[/img][/url]'
            ),
        ),
        (
            (ImageUrl('a.png'), ImageUrl('b.png'), ImageUrl('c.png')),
            2, ' | ', '\n-\n',
            (
                '[url=a.png][img]thumb_a.png[/img][/url] | [url=b.png][img]thumb_b.png[/img][/url]'
                '\n-\n'
                '[url=c.png][img]thumb_c.png[/img][/url]'
            ),
        ),
        (
            (ImageUrl('a.png'), ImageUrl('b.png'), ImageUrl('c.png')),
            3, '|', '-',
            (
                '[url=a.png][img]thumb_a.png[/img][/url]|[url=b.png][img]thumb_b.png[/img][/url]|[url=c.png][img]thumb_c.png[/img][/url]'
            ),
        ),
        (
            (ImageUrl('a.png'), ImageUrl('b.png'), ImageUrl('c.png'), ImageUrl('d.png')),
            2, '.', ':',
            (
                '[url=a.png][img]thumb_a.png[/img][/url].[url=b.png][img]thumb_b.png[/img][/url]'
                ':'
                '[url=c.png][img]thumb_c.png[/img][/url].[url=d.png][img]thumb_d.png[/img][/url]'
            ),
        ),
        (
            (ImageUrl('a.png'), ImageUrl('b.png'), ImageUrl('c.png'), ImageUrl('d.png'), ImageUrl('e.png')),
            2, '.', ':',
            (
                '[url=a.png][img]thumb_a.png[/img][/url].[url=b.png][img]thumb_b.png[/img][/url]'
                ':'
                '[url=c.png][img]thumb_c.png[/img][/url].[url=d.png][img]thumb_d.png[/img][/url]'
                ':'
                '[url=e.png][img]thumb_e.png[/img][/url]'
            ),
        ),
        (
            (ImageUrl('a.png'), ImageUrl('b.png'), ImageUrl('c.png'), ImageUrl('d.png'), ImageUrl('e.png')),
            3, '.', ':',
            (
                '[url=a.png][img]thumb_a.png[/img][/url].[url=b.png][img]thumb_b.png[/img][/url].[url=c.png][img]thumb_c.png[/img][/url]'
                ':'
                '[url=d.png][img]thumb_d.png[/img][/url].[url=e.png][img]thumb_e.png[/img][/url]'
            ),
        ),
        (
            (ImageUrl('a.png'), ImageUrl('b.png'), ImageUrl('c.png'), ImageUrl('d.png'), ImageUrl('e.png'), ImageUrl('f.png')),
            2, '.', ':',
            (
                '[url=a.png][img]thumb_a.png[/img][/url].[url=b.png][img]thumb_b.png[/img][/url]'
                ':'
                '[url=c.png][img]thumb_c.png[/img][/url].[url=d.png][img]thumb_d.png[/img][/url]'
                ':'
                '[url=e.png][img]thumb_e.png[/img][/url].[url=f.png][img]thumb_f.png[/img][/url]'
            ),
        ),
        (
            (ImageUrl('a.png'), ImageUrl('b.png'), ImageUrl('c.png'), ImageUrl('d.png'), ImageUrl('e.png'), ImageUrl('f.png')),
            3, '.', ':',
            (
                '[url=a.png][img]thumb_a.png[/img][/url].[url=b.png][img]thumb_b.png[/img][/url].[url=c.png][img]thumb_c.png[/img][/url]'
                ':'
                '[url=d.png][img]thumb_d.png[/img][/url].[url=e.png][img]thumb_e.png[/img][/url].[url=f.png][img]thumb_f.png[/img][/url]'
            ),
        ),
    ),
)
def test_make_screenshots_grid(screenshots, columns, horizontal_spacer, vertical_spacer, exp_bbcode, mocker):
    tracker_jobs = make_TestTrackerJobs(
        content_path='path/to/content',
        common_job_args={'home_directory': 'path/to/home', 'ignore_cache': 'mock bool'},
    )
    bbcode = tracker_jobs.make_screenshots_grid(
        screenshots,
        columns=columns,
        horizontal_spacer=horizontal_spacer,
        vertical_spacer=vertical_spacer,
    )
    assert bbcode == exp_bbcode

def test_make_screenshots_grid_without_thumbnail_url(mocker):
    tracker_jobs = make_TestTrackerJobs(
        content_path='path/to/content',
        common_job_args={'home_directory': 'path/to/home', 'ignore_cache': 'mock bool'},
    )
    screenshots = (ImageUrl('a.png'), ImageUrl('b.png', thumbnail=False), ImageUrl('c.png'))
    with pytest.raises(RuntimeError, match=r'^No thumbnail for b\.png$'):
        tracker_jobs.make_screenshots_grid(screenshots)



import os
from unittest.mock import AsyncMock, Mock, call

import pytest

from upsies import errors
from upsies.jobs.scene import SceneCheckJob
from upsies.utils.types import SceneCheckResult


@pytest.fixture
def make_SceneCheckJob(tmp_path):
    def make_SceneCheckJob(predb=None, force=None, content_path=tmp_path, ignore_cache=False):
        return SceneCheckJob(
            force=force,
            home_directory=tmp_path,
            cache_directory=tmp_path,
            ignore_cache=ignore_cache,
            content_path=content_path,
            predb=predb,
        )
    return make_SceneCheckJob


def test_cache_id(make_SceneCheckJob):
    job = make_SceneCheckJob(content_path='path/to/Foo/')
    assert job.cache_id is None


@pytest.mark.parametrize('path', ('', 'path/to'))
@pytest.mark.parametrize(
    argnames='content_path, exp_content_stem',
    argvalues=(
        ('Foo', 'Foo'),
        ('//Foo///', 'Foo'),
        ('Foo.mkv', 'Foo'),
    ),
)
def test_content_stem(path, content_path, exp_content_stem, make_SceneCheckJob):
    job = make_SceneCheckJob(content_path=os.path.join(path, content_path))
    assert job.content_stem == exp_content_stem


@pytest.mark.parametrize(
    argnames='kwargs, exp_attributes',
    argvalues=(
        ({}, {'_predetermined_result': None}),
        ({'force': None}, {'_predetermined_result': None}),
        ({'force': True}, {'_predetermined_result': True}),
        ({'force': 'false'}, {'_predetermined_result': True}),
        ({'force': False}, {'_predetermined_result': False}),
        ({'force': 0}, {'_predetermined_result': False}),
        ({'predb': 'mock predb'}, {'_predb': 'mock predb'}),
    ),
    ids=lambda v: str(v),
)
def test_initialize(kwargs, exp_attributes, make_SceneCheckJob, mocker, event_loop):
    job = make_SceneCheckJob(**kwargs)
    for name, value in exp_attributes.items():
        assert getattr(job, name) == value


@pytest.mark.parametrize(
    argnames='exc, exp_exc, exp_cbs, exp_errors, exp_warnings',
    argvalues=(
        (
            None,
            None,
            [],
            (),
            (),
        ),
        (
            errors.SceneError('Foo'),
            None,
            [],
            (errors.SceneError('Foo'),),
            (),
        ),
        (
            errors.RequestError('Bar'),
            None,
            [call(SceneCheckResult.unknown)],
            (),
            ('Bar',),
        ),
    ),
    ids=lambda v: str(v),
)
@pytest.mark.asyncio
async def test_catch_errors(exc, exp_exc, exp_cbs, exp_errors, exp_warnings, make_SceneCheckJob, mocker, event_loop):
    async def coro():
        if exc:
            raise exc

    ask_is_scene_release_cb = Mock(side_effect=lambda scene_check_result: job.finish())

    job = make_SceneCheckJob()
    job.signal.register('ask_is_scene_release', ask_is_scene_release_cb)

    await job._catch_errors(coro())
    if not exc:
        event_loop.call_soon(job.finish)

    assert job.errors == exp_errors
    assert job.warnings == exp_warnings

    assert ask_is_scene_release_cb.call_args_list == exp_cbs


@pytest.mark.parametrize(
    argnames='predetermined_result, exp_catch_errors_called, exp_verify_calls, exp_finalize_calls',
    argvalues=(
        (None, True, [call()], []),
        (True, False, [], [call(SceneCheckResult.true)]),
        (False, False, [], [call(SceneCheckResult.false)]),
        ('asdf', False, [], []),
    ),
)
@pytest.mark.asyncio
async def test_execute(predetermined_result,
                       exp_catch_errors_called, exp_verify_calls, exp_finalize_calls,
                       make_SceneCheckJob, mocker):
    mocker.patch('upsies.jobs.scene.SceneCheckJob._catch_errors', AsyncMock())
    mocker.patch('upsies.jobs.scene.SceneCheckJob._verify', Mock())
    mocker.patch('upsies.jobs.scene.SceneCheckJob.finalize', Mock())
    job = make_SceneCheckJob()
    job._predetermined_result = predetermined_result

    if not exp_catch_errors_called and not exp_verify_calls and not exp_finalize_calls:
        with pytest.raises(RuntimeError, match=rf'^Unexpected predetermined_result: {predetermined_result!r}'):
            job.execute()
    else:
        job.execute()
        await job.await_tasks()
        assert job.finalize.call_args_list == exp_finalize_calls
        assert job._verify.call_args_list == exp_verify_calls
        if exp_catch_errors_called:
            assert job._catch_errors.call_args_list == [call(job._verify.return_value)]
        else:
            assert job._catch_errors.call_args_list == []


@pytest.mark.parametrize(
    argnames='scene_check_result, exp_finalize_called',
    argvalues=(
        (SceneCheckResult.false, True),
        (SceneCheckResult.unknown, False),
        (SceneCheckResult.true, False),
    ),
)
@pytest.mark.asyncio
async def test_verify_finalizes_early_if_not_scene(scene_check_result, exp_finalize_called, make_SceneCheckJob, mocker):
    job = make_SceneCheckJob(
        content_path='path/to/foo',
        predb=Mock(
            search=AsyncMock(return_value=('first mock release', 'second mock release')),
            is_scene_release=AsyncMock(return_value=scene_check_result),
            verify_release=AsyncMock(),
        ),
    )
    is_mixed_season_pack_mock = mocker.patch('upsies.utils.predbs.is_mixed_season_pack', return_value=True)
    mocker.patch.object(job, '_verify_release', AsyncMock(
        return_value=('mock is_scene_release', 'mock exceptions'),
    ))
    mocker.patch.object(job, 'finalize', Mock())

    ask_release_name_cb = Mock()
    job.signal.register('ask_release_name', ask_release_name_cb)

    await job._verify()

    assert job._predb.is_scene_release.call_args_list == [call('path/to/foo')]
    if exp_finalize_called:
        assert job.finalize.call_args_list == [call(SceneCheckResult.false)]
        assert is_mixed_season_pack_mock.call_args_list == []
        assert ask_release_name_cb.call_args_list == []
        assert job._predb.search.call_args_list == []
        assert job._verify_release.call_args_list == []
    else:
        assert job.finalize.call_args_list == []
        assert is_mixed_season_pack_mock.call_args_list == [call('path/to/foo')]
        assert ask_release_name_cb.call_args_list == []
        assert job._predb.search.call_args_list == []
        assert job._verify_release.call_args_list == [call()]

@pytest.mark.asyncio
async def test_verify_verifies_release_if_it_is_mixed(make_SceneCheckJob, mocker):
    job = make_SceneCheckJob(
        content_path='path/to/foo',
        predb=Mock(
            search=AsyncMock(return_value=('first mock release', 'second mock release')),
            is_scene_release=AsyncMock(return_value=SceneCheckResult.true),
            verify_release=AsyncMock(),
        ),
    )
    is_mixed_season_pack_mock = mocker.patch('upsies.utils.predbs.is_mixed_season_pack', return_value=True)
    mocker.patch.object(job, '_verify_release', AsyncMock(
        return_value=('mock is_scene_release', 'mock exceptions'),
    ))
    mocker.patch.object(job, 'finalize', Mock())

    ask_release_name_cb = Mock()
    job.signal.register('ask_release_name', ask_release_name_cb)

    await job._verify()

    assert job._predb.is_scene_release.call_args_list == [call('path/to/foo')]
    assert job.finalize.call_args_list == []
    assert is_mixed_season_pack_mock.call_args_list == [call('path/to/foo')]
    assert ask_release_name_cb.call_args_list == []
    assert job._predb.search.call_args_list == []
    assert job._verify_release.call_args_list == [call()]

@pytest.mark.parametrize('extension', (None, 'mkv'))
@pytest.mark.parametrize(
    argnames='stem, search_results, exp_action',
    argvalues=(
        (
            'Foo.2003.720p.BluRay.x264-ASDF',
            ('Foo.2003.720p.BluRay.x264-ASDF', 'Foo.2003.720p.REPACK.BluRay.x264-ASDF'),
            'autoselect',
        ),
        (
            'Foo.2003.720p.REPACK.BluRay.x264-ASDF',
            ('Foo.2003.720p.BluRay.x264-ASDF', 'Foo.2003.720p.REPACK.BluRay.x264-ASDF'),
            'autoselect',
        ),
        (
            'Foo.2003.720p.BONUS.BluRay.x264-ASDF',
            ('Foo.2003.720p.BluRay.x264-ASDF', 'Foo.2003.720p.REPACK.BluRay.x264-ASDF'),
            'ask_user',
        ),
    ),
    ids=lambda v: repr(v),
)
@pytest.mark.asyncio
async def test_verify_autoselects_correctly_named_release_from_multiple_search_results(extension,
                                                                                       stem, search_results, exp_action,
                                                                                       make_SceneCheckJob, mocker):
    content_path = f'path/to/{stem}'
    if extension:
        content_path += f'.{extension}'

    job = make_SceneCheckJob(
        content_path=content_path,
        predb=Mock(
            search=AsyncMock(return_value=search_results),
            is_scene_release=AsyncMock(return_value=SceneCheckResult.true),
            verify_release=AsyncMock(),
        ),
    )
    is_mixed_season_pack_mock = mocker.patch('upsies.utils.predbs.is_mixed_season_pack', return_value=False)
    mocker.patch.object(job, '_verify_release', AsyncMock(
        return_value=('mock is_scene_release', 'mock exceptions')
    ))
    mocker.patch.object(job, 'finalize', Mock())

    ask_release_name_cb = Mock()
    job.signal.register('ask_release_name', ask_release_name_cb)

    await job._verify()

    assert job._predb.is_scene_release.call_args_list == [call(content_path)]
    assert job.finalize.call_args_list == []
    assert is_mixed_season_pack_mock.call_args_list == [call(content_path)]
    assert job._predb.search.call_args_list == [call(query=content_path, only_existing_releases=False)]
    if exp_action == 'autoselect':
        assert ask_release_name_cb.call_args_list == []
        assert job._verify_release.call_args_list == [call()]
    elif exp_action == 'ask_user':
        assert ask_release_name_cb.call_args_list == [call(search_results)]
        assert job._verify_release.call_args_list == []
    else:
        raise RuntimeError(f'Unexpected exp_action: {exp_action!r}')

@pytest.mark.asyncio
async def test_verify_asks_user_if_there_are_multiple_search_results(make_SceneCheckJob, mocker):
    job = make_SceneCheckJob(
        content_path='path/to/foo',
        predb=Mock(
            search=AsyncMock(return_value=('first mock release', 'second mock release')),
            is_scene_release=AsyncMock(return_value=SceneCheckResult.true),
            verify_release=AsyncMock(),
        ),
    )
    is_mixed_season_pack_mock = mocker.patch('upsies.utils.predbs.is_mixed_season_pack', return_value=False)
    mocker.patch.object(job, '_verify_release', AsyncMock(
        return_value=('mock is_scene_release', 'mock exceptions')
    ))
    mocker.patch.object(job, 'finalize', Mock())

    ask_release_name_cb = Mock()
    job.signal.register('ask_release_name', ask_release_name_cb)

    await job._verify()

    assert job._predb.is_scene_release.call_args_list == [call('path/to/foo')]
    assert job.finalize.call_args_list == []
    assert is_mixed_season_pack_mock.call_args_list == [call('path/to/foo')]
    assert ask_release_name_cb.call_args_list == [call(('first mock release', 'second mock release'))]
    assert job._predb.search.call_args_list == [call(query='path/to/foo', only_existing_releases=False)]
    assert job._verify_release.call_args_list == []

@pytest.mark.asyncio
async def test_verify_verifies_release_if_there_is_one_search_result(make_SceneCheckJob, mocker):
    job = make_SceneCheckJob(
        content_path='path/to/foo',
        predb=Mock(
            search=AsyncMock(return_value=('single mock release',)),
            is_scene_release=AsyncMock(return_value=SceneCheckResult.true),
            verify_release=AsyncMock(),
        ),
    )
    is_mixed_season_pack_mock = mocker.patch('upsies.utils.predbs.is_mixed_season_pack', return_value=False)
    mocker.patch.object(job, '_verify_release', AsyncMock(
        return_value=('mock is_scene_release', 'mock exceptions')
    ))
    mocker.patch.object(job, 'finalize', Mock())

    ask_release_name_cb = Mock()
    job.signal.register('ask_release_name', ask_release_name_cb)

    await job._verify()

    assert job._predb.is_scene_release.call_args_list == [call('path/to/foo')]
    assert job.finalize.call_args_list == []
    assert is_mixed_season_pack_mock.call_args_list == [call('path/to/foo')]
    assert ask_release_name_cb.call_args_list == []
    assert job._predb.search.call_args_list == [call(query='path/to/foo', only_existing_releases=False)]
    assert job._verify_release.call_args_list == [call()]


@pytest.mark.asyncio
async def test_user_selected_release_name_with_release_name(make_SceneCheckJob, mocker):
    job = make_SceneCheckJob()
    mocker.patch.object(job, '_verify_release')
    mocker.patch.object(job, '_handle_scene_check_result')

    job.user_selected_release_name('mock.release.name')

    await job.await_tasks()
    assert job._verify_release.call_args_list == [call('mock.release.name')]
    assert job._handle_scene_check_result.call_args_list == []

@pytest.mark.asyncio
async def test_user_selected_release_name_without_release_name(make_SceneCheckJob, mocker):
    job = make_SceneCheckJob()
    mocker.patch.object(job, '_verify_release')
    mocker.patch.object(job, '_handle_scene_check_result')

    job.user_selected_release_name(None)

    await job.await_tasks()
    assert job._verify_release.call_args_list == []
    assert job._handle_scene_check_result.call_args_list == [call(SceneCheckResult.false)]


@pytest.mark.asyncio
async def test_verify_release(make_SceneCheckJob, mocker):
    job = make_SceneCheckJob(
        predb=Mock(
            verify_release=AsyncMock(
                return_value=(SceneCheckResult.true, ('error1', 'error2')),
            ),
        ),
    )
    mocker.patch.object(job, '_handle_scene_check_result')

    await job._verify_release('mock.release.name')

    assert job._predb.verify_release.call_args_list == [
        call(job._content_path, 'mock.release.name'),
    ]
    assert job._handle_scene_check_result.call_args_list == [
        call(SceneCheckResult.true, ('error1', 'error2')),
    ]


def test_handle_scene_check_result_handles_SceneErrors(make_SceneCheckJob, mocker):
    mocker.patch('upsies.jobs.scene.SceneCheckJob.finalize')
    ask_is_scene_release = Mock()
    job = make_SceneCheckJob()
    job.signal.register('ask_is_scene_release', ask_is_scene_release)
    job._handle_scene_check_result(
        'mock scene check result',
        exceptions=(
            errors.SceneError('foo'),
            errors.SceneRenamedError(original_name='bar', existing_name='Bar'),
            errors.SceneMissingInfoError('burr'),
            errors.SceneFileSizeError('baz', original_size=123, existing_size=456),
        ),
    )
    assert job.warnings == (str(errors.SceneMissingInfoError('burr')),)
    assert job.errors == (
        errors.SceneError('foo'),
        errors.SceneRenamedError(original_name='bar', existing_name='Bar'),
        errors.SceneFileSizeError('baz', original_size=123, existing_size=456),
    )
    assert ask_is_scene_release.call_args_list == []
    assert job.finalize.call_args_list == []
    assert job.is_finished

@pytest.mark.parametrize('is_scene_release', (SceneCheckResult.true, SceneCheckResult.false))
def test_handle_scene_check_result_handles_definite_scene_check_result(is_scene_release, make_SceneCheckJob, mocker):
    mocker.patch('upsies.jobs.scene.SceneCheckJob.finalize')
    ask_is_scene_release = Mock()
    job = make_SceneCheckJob()
    job.signal.register('ask_is_scene_release', ask_is_scene_release)
    job._handle_scene_check_result(is_scene_release, exceptions=())
    assert job.errors == ()
    assert job.warnings == ()
    assert ask_is_scene_release.call_args_list == []
    assert job.finalize.call_args_list == [call(is_scene_release)]
    assert not job.is_finished

@pytest.mark.parametrize('is_scene_release', (SceneCheckResult.unknown,))
def test_handle_scene_check_result_triggers_dialog_if_no_errors(is_scene_release, make_SceneCheckJob, mocker):
    mocker.patch('upsies.jobs.scene.SceneCheckJob.finalize')
    ask_is_scene_release = Mock()
    job = make_SceneCheckJob()
    job.signal.register('ask_is_scene_release', ask_is_scene_release)
    job._handle_scene_check_result('mock scene check result', ())
    assert job.errors == ()
    assert job.warnings == ()
    assert ask_is_scene_release.call_args_list == [
        call('mock scene check result'),
    ]
    assert job.finalize.call_args_list == []
    assert not job.is_finished


@pytest.mark.parametrize(
    argnames='is_scene_release, exp_msg',
    argvalues=(
        (SceneCheckResult.true, 'Scene release'),
        (SceneCheckResult.false, 'Not a scene release'),
        (SceneCheckResult.unknown, 'May be a scene release'),
    ),
    ids=lambda v: str(v),
)
def test_finalize(is_scene_release, exp_msg, make_SceneCheckJob, mocker):
    cb = Mock()
    job = make_SceneCheckJob()
    job.signal.register('checked', cb)
    assert job.is_scene_release is None
    job.finalize(is_scene_release)
    assert job.output == (exp_msg,)
    assert cb.call_args_list == [call(is_scene_release)]
    assert job.is_scene_release is is_scene_release
    assert job.is_finished


@pytest.mark.parametrize(
    argnames='is_scene_release',
    argvalues=(
        SceneCheckResult.true,
        SceneCheckResult.false,
        SceneCheckResult.unknown,
    ),
    ids=lambda v: str(v),
)
@pytest.mark.asyncio
async def test_is_scene_release_property_is_not_cached(is_scene_release, make_SceneCheckJob, mocker):
    mocker.patch('upsies.jobs.scene.SceneCheckJob._verify', AsyncMock())

    # Create job output
    job = make_SceneCheckJob(ignore_cache=False)
    assert job.is_scene_release is None
    job.start()
    await job.await_tasks()
    job.finalize(is_scene_release)
    assert job.is_scene_release is is_scene_release

    # Run the same job again and expect no cached output
    job = make_SceneCheckJob(ignore_cache=False)
    job.start()
    await job.await_tasks()
    assert job.is_scene_release is None



from unittest.mock import Mock, call

import pytest

from upsies.utils import imghosts


def test_imghosts(mocker):
    existing_imghosts = (Mock(), Mock(), Mock())
    submodules_mock = mocker.patch('upsies.utils.imghosts.submodules')
    subclasses_mock = mocker.patch('upsies.utils.imghosts.subclasses', return_value=existing_imghosts)
    assert imghosts.imghosts() == existing_imghosts
    assert submodules_mock.call_args_list == [call('upsies.utils.imghosts')]
    assert subclasses_mock.call_args_list == [call(imghosts.base.ImageHostBase, submodules_mock.return_value)]


@pytest.mark.parametrize('name', ('foo', 'bar', 'baz'))
@pytest.mark.parametrize(
    argnames='options, cache_directory, exp_calls',
    argvalues=(
        (None, None, [call(options=None, cache_directory=None)]),
        ({'this': 'that'}, None, [call(options={'this': 'that'}, cache_directory=None)]),
        (None, 'path/to/cache', [call(options=None, cache_directory='path/to/cache')]),
        ({'this': 'that'}, 'path/to/cache', [call(options={'this': 'that'}, cache_directory='path/to/cache')]),
    ),
)
def test_imghost_returns_ImageHostBase_instance(name, options, cache_directory, exp_calls, mocker):
    existing_imghosts = {
        'foo': Mock(),
        'bar': Mock(),
        'baz': Mock(),
    }
    existing_imghosts['foo'].configure_mock(name='foo')
    existing_imghosts['bar'].configure_mock(name='bar')
    existing_imghosts['baz'].configure_mock(name='baz')
    mocker.patch('upsies.utils.imghosts.imghosts', return_value=existing_imghosts.values())

    kwargs = {}
    if options:
        kwargs['options'] = options
    if cache_directory:
        kwargs['cache_directory'] = cache_directory
    return_value = imghosts.imghost(name, **kwargs)
    assert return_value == existing_imghosts[name].return_value
    assert existing_imghosts[name].call_args_list == exp_calls

def test_imghost_fails_to_find_imghost(mocker):
    existing_imghosts = (Mock(), Mock(), Mock())
    existing_imghosts[0].configure_mock(name='foo')
    existing_imghosts[1].configure_mock(name='bar')
    existing_imghosts[2].configure_mock(name='baz')
    mocker.patch('upsies.utils.imghosts.imghosts', return_value=existing_imghosts)
    with pytest.raises(ValueError, match='^Unsupported image hosting service: bam$'):
        imghosts.imghost('bam')
    for ih in existing_imghosts:
        assert ih.call_args_list == []


def test_imghost_names(mocker):
    existing_imghosts = (Mock(), Mock(), Mock())
    existing_imghosts[0].configure_mock(name='FOO')
    existing_imghosts[1].configure_mock(name='bar')
    existing_imghosts[2].configure_mock(name='Baz')
    mocker.patch('upsies.utils.imghosts.imghosts', return_value=existing_imghosts)
    assert imghosts.imghost_names() == ['bar', 'Baz', 'FOO']



import re

import pytest

from upsies import errors


def error_classes():
    clses = []
    for name in dir(errors):
        if not name.startswith('_'):
            cls = getattr(errors, name)
            if isinstance(cls, type) and issubclass(cls, Exception):
                clses.append(cls)
    return clses


@pytest.mark.parametrize(
    argnames='cls',
    argvalues=error_classes(),
)
def test_equality(cls):
    try:
        assert cls('foo') == cls('foo')
        assert cls('foo') != cls('bar')
        assert cls('foo') != ValueError('foo')
    # Some exceptions require more arguments
    except TypeError:
        pass


def test_RequestError_url():
    e = errors.RequestError('foo', url='http://foo')
    assert e.url == 'http://foo'

def test_RequestError_headers():
    e = errors.RequestError('foo', headers={'a': 1, 'b': 2})
    assert e.headers == {'a': 1, 'b': 2}

def test_RequestError_status_code():
    e = errors.RequestError('foo', status_code=123)
    assert e.status_code == 123

def test_RequestError_text():
    e = errors.RequestError('foo', text='Error 404')
    assert e.text == 'Error 404'

@pytest.mark.parametrize(
    argnames='text, default, exp_result',
    argvalues=(
        ('foo', errors._NO_DEFAULT_VALUE, errors.RequestError('Malformed JSON: foo: Expecting value: line 1 column 1 (char 0)')),
        ('foo', None, None),
        ('foo', 'asdf', 'asdf'),
        ('"foo"', None, 'foo'),
        ('["foo", "bar"]', 'asdf', ['foo', 'bar']),
        ('["foo", "bar", baz]', 'asdf', 'asdf'),
    ),
)
def test_RequestError_json(text, default, exp_result):
    e = errors.RequestError('foo', text=text)
    if isinstance(exp_result, Exception):
        with pytest.raises(type(exp_result), match=rf'^{re.escape(str(exp_result))}$'):
            e.json(default=default)
    else:
        assert e.json(default=default) == exp_result


def test_SubprocessError():
    e = TypeError('foo')
    traceback = 'mock traceback'
    subproc_e = errors.SubprocessError(e, traceback)
    assert subproc_e.original_traceback == 'Subprocess traceback:\nmock traceback'


def test_SceneRenamedError_name_attributes():
    e = errors.SceneRenamedError(original_name='foo', existing_name='bar')
    assert e.original_name == 'foo'
    assert e.existing_name == 'bar'


def test_SceneFileSizeError():
    e = errors.SceneFileSizeError(
        filename='foo',
        original_size=123,
        existing_size=124,
    )
    assert e.filename == 'foo'
    assert e.original_size == 123
    assert e.existing_size == 124


def test_SceneMissingInfoError():
    e = errors.SceneMissingInfoError('foo.mkv')
    assert str(e) == 'Missing information: foo.mkv'


@pytest.mark.parametrize(
    argnames='files, exp_msg',
    argvalues=(
        (
            [],
            (
                'Potential duplicate files found'
            ),
        ),
        (
            ['a.mkv'],
            (
                '1 potential duplicate file found:\n'
                ' - a.mkv'
            ),
        ),
        (
            ['a.mkv', 'b.mkv'],
            (
                '2 potential duplicate files found:\n'
                ' - a.mkv\n'
                ' - b.mkv'
            ),
        ),
    ),
    ids=lambda v: repr(v),
)
def test_FoundDupeError(files, exp_msg):
    e = errors.FoundDupeError(files)
    assert str(e) == exp_msg


def test_AnnounceUrlNotsetError():
    tracker = 'mock tracker'
    e = errors.AnnounceUrlNotSetError(tracker=tracker)
    assert str(e) == 'Announce URL is not set'
    assert e.tracker is tracker



from unittest.mock import AsyncMock, Mock, call

import pytest

from upsies import errors
from upsies.jobs.torrent import AddTorrentJob


@pytest.fixture
def btclient():
    class MockBtClient():
        name = 'mocksy'
        label = 'Mocksy'
        add_torrent = AsyncMock()

        def __repr__(self):
            return '<BtClient instance>'

    return MockBtClient()


@pytest.fixture
async def make_AddTorrentJob(tmp_path, btclient):
    def make_AddTorrentJob(download_path=tmp_path, check_after_add='<check_after_add>', torrents=()):
        return AddTorrentJob(
            home_directory=tmp_path,
            cache_directory=tmp_path,
            ignore_cache=False,
            btclient=btclient,
            enqueue=torrents,
        )
    return make_AddTorrentJob


def test_cache_id(make_AddTorrentJob):
    job = make_AddTorrentJob()
    assert job.cache_id is None


@pytest.mark.parametrize(
    argnames='signal, args, exp_info',
    argvalues=(
        ('adding', ('path/to.torrent',), 'Adding to.torrent'),
        ('added', ('ignored argument',), ''),
        ('finished', ('ignored argument',), ''),
        ('error', ('ignored argument',), ''),
    ),
    ids=lambda v: repr(v),
)
@pytest.mark.asyncio
async def test_info_property_is_set(signal, args, exp_info, make_AddTorrentJob):
    job = make_AddTorrentJob()
    job.info = 'foo!'
    assert job.info == 'foo!'
    job.signal.emit(signal, *args)
    assert job.info == exp_info


@pytest.mark.parametrize(
    argnames='torrent_path, add_torrent_result, exp_calls',
    argvalues=(
        pytest.param(
            'my.torrent',
            errors.TorrentAddError('bad torrent'),
            [
                call.emit('adding', 'my.torrent'),
                call.add_torrent('my.torrent'),
                call.error(errors.TorrentAddError('bad torrent')),
            ],
            id='Adding fails',
        ),
        pytest.param(
            'my.torrent',
            'd34db33f',
            [
                call.emit('adding', 'my.torrent'),
                call.add_torrent('my.torrent'),
                call.send('d34db33f'),
                call.emit('added', 'd34db33f'),
            ],
            id='Adding succeeds',
        ),
    ),
)
@pytest.mark.asyncio
async def test_handle_input(torrent_path, add_torrent_result, exp_calls, make_AddTorrentJob, btclient, mocker):
    job = make_AddTorrentJob()

    mocks = Mock()
    mocks.attach_mock(
        (
            mocker.patch.object(job._btclient, 'add_torrent', side_effect=add_torrent_result)
            if isinstance(add_torrent_result, Exception) else
            mocker.patch.object(job._btclient, 'add_torrent', return_value=add_torrent_result)
        ),
        'add_torrent',
    )
    mocks.attach_mock(mocker.patch.object(job.signal, 'emit'), 'emit')
    mocks.attach_mock(mocker.patch.object(job, 'error'), 'error')
    mocks.attach_mock(mocker.patch.object(job, 'send'), 'send')

    await job.handle_input(torrent_path)
    assert mocks.mock_calls == exp_calls



import os
import sys

sys.path.append(os.path.abspath("./_ext"))

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.intersphinx',
    'sphinx.ext.viewcode',
    'cli_reference',
]

autosummary_generate = True
html_show_sourcelink = False  # Don't show links to rST code

# This is ignored for some reason?
autodoc_member_order = 'bysource'

templates_path = ['_templates']

intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "aiobtclientapi": ("https://aiobtclientapi.readthedocs.io/en/stable/", None),
}



"""
Abstract base class for all tracker-specific stuff
"""

import abc
import asyncio
import functools

from ... import errors, utils
from ._howto import Howto

import logging  # isort:skip
_log = logging.getLogger(__name__)


class TrackerBase(abc.ABC):
    """
    Base class for tracker-specific operations, e.g. uploading

    :param options: User configuration options for this tracker,
        e.g. authentication details, announce URL, etc
    :type options: :class:`dict`-like
    """

    @property
    @abc.abstractmethod
    def TrackerJobs(self):
        """Subclass of :class:`TrackerJobsBase`"""

    @property
    @abc.abstractmethod
    def TrackerConfig(self):
        """Subclass of :class:`TrackerConfigBase`"""

    def __init__(self, options=None):
        self._options = options or {}
        self._signal = utils.signal.Signal('warning', 'error', 'exception')

    @property
    @abc.abstractmethod
    def name(self):
        """Lower-case tracker name abbreviation for internal use"""

    @property
    @abc.abstractmethod
    def label(self):
        """User-facing tracker name abbreviation"""

    setup_howto_template = 'Nobody has written a setup howto yet.'
    """
    Step-by-step guide that explains how to make your first upload

    .. note:: This MUST be a class attribute and not a property.
    """

    @classmethod
    def generate_setup_howto(cls):
        # `howto` is now in the local namespace (see `locals()`) and
        # `evalute_fstring()` can access its properties to replace, for example,
        # "{howto.current}" in `setup_howto_template` with the current section
        # number.
        howto = Howto(tracker_cls=cls)  # noqa: F841 local variable 'howto' is assigned to but never used
        return utils.string.evaluate_fstring(cls.setup_howto_template)

    @property
    def options(self):
        """
        Configuration options provided by the user

        This is the :class:`dict`-like object from the initialization argument
        of the same name.
        """
        return self._options

    @functools.cached_property
    def _tasks(self):
        return []

    def attach_task(self, coro, callback=None):
        """
        Run awaitable `coro` in background task

        :param coro: Any awaitable

        :param callback: Callable that is called with the return value or
            exception from `coro`
        """
        task = utils.run_task(coro, callback=callback)
        self._tasks.append(task)
        return task

    async def await_tasks(self):
        """Wait for all awaitables passed to :meth:`attach_task`"""
        for task in self._tasks:
            await task

    async def login(self):
        """
        Start user session

        Set :attr:`is_logged_in` to `True` or `False`.

        :raise errors.RequestError: on failure
        """
        async with self._session_lock:
            if not self.is_logged_in:
                try:
                    await self._login()
                except errors.RequestError:
                    self._is_logged_in = False
                    raise
                else:
                    self._is_logged_in = True

    async def logout(self):
        """
        End user session

        Always set :attr:`is_logged_in` to `False`, even on failure.

        :raise errors.RequestError: on failure
        """
        async with self._session_lock:
            if self.is_logged_in:
                try:
                    await self._logout()
                finally:
                    self._is_logged_in = False

    @abc.abstractmethod
    async def _login(self):
        """
        Start user session

        :raise errors.RequestError: on any kind of failure
        """

    @abc.abstractmethod
    async def _logout(self):
        """
        End user session

        :raise errors.RequestError: on any kind of failure
        """

    @functools.cached_property
    def _session_lock(self):
        # Prevent multiple simultaneous login/logout attempts
        return asyncio.Lock()

    @property
    def is_logged_in(self):
        """Whether a user session is active"""
        return getattr(self, '_is_logged_in', False)

    @abc.abstractmethod
    async def get_announce_url(self):
        """Get announce URL from tracker website"""

    calculate_piece_size = None
    """
    :class:`staticmethod` that takes a torrent's content size and returns
    the corresponding piece size

    If this is `None`, the default implementation is used.
    """

    calculate_piece_size_min_max = None
    """
    :class:`staticmethod` that takes a torrent's content size and returns
    the corresponding allowed minimum and maximum piece sizes

    If this is `None`, the default implementation is used.
    """

    @abc.abstractmethod
    async def upload(self, tracker_jobs):
        """
        Upload torrent and other metadata from jobs

        :param TrackerJobsBase tracker_jobs: :attr:`TrackerJobs` instance
        """

    @property
    def signal(self):
        """
        :class:`~.signal.Signal` instance with the signals ``warning``, ``error``
        and ``exception``
        """
        return self._signal

    def warn(self, warning):
        """
        Emit ``warning`` signal (see :attr:`signal`)

        Emit a warning for any non-critical issue that the user can choose to
        ignore or fix.
        """
        self.signal.emit('warning', warning)

    def error(self, error):
        """
        Emit ``error`` signal (see :attr:`signal`)

        Emit an error for any critical but expected issue that can't be
        recovered from (e.g. I/O error).
        """
        self.signal.emit('error', error)

    def exception(self, exception):
        """
        Emit ``exception`` signal (see :attr:`signal`)

        Emit an exception for any critical and unexpected issue that should be
        reported as a bug.
        """
        self.signal.emit('exception', exception)



"""
Execute external commands
"""

import os

from .. import errors
from ..utils import LazyModule

import logging  # isort:skip
_log = logging.getLogger(__name__)

subprocess = LazyModule(module='subprocess', namespace=globals())
shlex = LazyModule(module='shlex', namespace=globals())

_command_output_cache = {}


def run(argv, ignore_errors=False, join_stderr=False, cache=False):
    """
    Execute command in subprocess

    :param argv: Command to execute
    :type argv: list of str
    :param bool ignore_errors: Do not raise :class:`.ProcessError` if stderr is
        non-empty
    :param bool join_stderr: Redirect stderr to stdout
    :param bool cache: Cache output based on `argv`

    :raise DependencyError: if the command fails to execute
    :raise ProcessError: if stdout is not empty and `ignore_errors` is `False`

    :return: Output from process
    :rtype: str
    """
    argv = tuple(str(arg) for arg in argv)
    if cache and argv in _command_output_cache:
        stdout, stderr = _command_output_cache[argv]
    else:
        fh_stdout = subprocess.PIPE
        if join_stderr:
            fh_stderr = subprocess.STDOUT
        else:
            fh_stderr = subprocess.PIPE
        try:
            _log.debug('Running: %s', ' '.join(shlex.quote(arg) for arg in argv))
            proc = subprocess.run(
                argv,
                shell=False,
                encoding='utf-8',
                stdout=fh_stdout,
                stderr=fh_stderr,
                stdin=subprocess.PIPE,
            )
        except OSError:
            raise errors.DependencyError(f'Missing dependency: {os.path.basename(argv[0])}')
        else:
            stdout, stderr = proc.stdout, proc.stderr
            if cache:
                _command_output_cache[argv] = (stdout, stderr)
    if stderr and not ignore_errors:
        raise errors.ProcessError(stderr)
    return stdout



import copy
import datetime
import errno
import os
import re
from types import SimpleNamespace
from unittest.mock import Mock, call, patch

import pytest
import torf

from upsies import __project_name__, __version__, errors
from upsies.utils import torrent, types


@pytest.mark.parametrize(
    argnames='announce, source, exp_error',
    argvalues=(
        ('http://foo', '', 'Source is empty'),
        ('', 'ASDF', 'Announce URL is empty'),
        ('', '', 'Announce URL is empty'),
    ),
)
def test_create_validates_arguments(announce, source, exp_error, mocker):
    mocker.patch('upsies.utils.torrent._get_torrent')
    mocker.patch('upsies.utils.torrent.Files')
    mocker.patch('upsies.utils.torrent._find_hashes')
    mocker.patch('upsies.utils.torrent._generate_hashes')
    mocker.patch('upsies.utils.torrent._store_generic_torrent')

    with pytest.raises(errors.TorrentCreateError, match=rf'^{re.escape(exp_error)}$'):
        torrent.create(
            content_path='/path/to/content',
            torrent_path='/path/to/content.torrent',
            announce=announce,
            source=source,
            init_callback=Mock(),
            progress_callback=Mock(),
            exclude=('a', 'b', 'c'),
        )
    assert torrent._get_torrent.call_args_list == []
    assert torrent.Files.call_args_list == []
    assert torrent._find_hashes.call_args_list == []
    assert torrent._generate_hashes.call_args_list == []
    assert torrent._store_generic_torrent.call_args_list == []


@pytest.mark.parametrize(
    argnames='piece_size_min_max_calculator, exp_piece_size_min, exp_piece_size_max',
    argvalues=(
        pytest.param(Mock(return_value=(123, 456)), 123, 456, id='Custom piece size min/max'),
        pytest.param(None, None, None, id='Default piece size min/max'),
    ),
)
@pytest.mark.parametrize(
    argnames='piece_size_calculator, exp_piece_size',
    argvalues=(
        pytest.param(Mock(return_value=456), 456, id='Custom piece size'),
        pytest.param(None, None, id='Default piece size'),
    ),
)
@pytest.mark.parametrize(
    argnames='init_cb_cancels, find_hashes_return_value, generate_hashes_cancels',
    argvalues=(
        pytest.param(False, False, False, id='nothing cancels'),
        pytest.param(True, False, False, id='init_callback cancels'),
        pytest.param(False, True, False, id='find_hashes cancels'),
        pytest.param(False, torrent.SKIP_SEARCHING, False, id='find_hashes skips searching'),
        pytest.param(False, False, True, id='generate_hashes cancels'),
    ),
)
@pytest.mark.parametrize(
    argnames='use_cache, find_hashes_succeeds',
    argvalues=(
        pytest.param(True, True, id='find hash successfully'),
        pytest.param(False, False, id='do not try to find hash'),
    ),
)
def test_create(piece_size_min_max_calculator, exp_piece_size_min, exp_piece_size_max,
                piece_size_calculator, exp_piece_size,
                init_cb_cancels, find_hashes_return_value, generate_hashes_cancels,
                use_cache, find_hashes_succeeds,
                mocker):
    def find_hashes_side_effect(**_):
        mock_torrent.is_ready = find_hashes_succeeds is True
        return find_hashes_return_value

    mocks = Mock()
    mock_torrent = Mock(is_ready=False)
    mocks.attach_mock(mocker.patch('upsies.utils.torrent._get_exclude_regexs', return_value=mock_torrent), '_get_exclude_regexs')
    mocks.attach_mock(mocker.patch('upsies.utils.torrent._get_torrent', return_value=mock_torrent), '_get_torrent')
    if piece_size_min_max_calculator:
        mocks.attach_mock(piece_size_min_max_calculator, 'piece_size_min_max_calculator')
    if piece_size_calculator:
        mocks.attach_mock(piece_size_calculator, 'piece_size_calculator')
    mocks.attach_mock(mocker.patch('upsies.utils.torrent.Files'), 'Files')
    mocks.attach_mock(mocker.patch('upsies.utils.torrent._find_hashes', side_effect=find_hashes_side_effect), '_find_hashes')
    mocks.attach_mock(mocker.patch('upsies.utils.torrent._generate_hashes', return_value=generate_hashes_cancels), '_generate_hashes')
    mocks.attach_mock(mocker.patch('upsies.utils.torrent._store_generic_torrent'), '_store_generic_torrent')
    mocks.attach_mock(mocker.patch('upsies.utils.torrent._write_torrent_path'), '_write_torrent_path')
    mocks.attach_mock(Mock(return_value=init_cb_cancels), 'init_callback')
    mocks.attach_mock(Mock(), 'progress_callback')

    torrent_path = '/path/to/content.torrent'
    content_path = '/path/to/content'
    exclude = ('a', 'b', 'c')
    randomize_infohash = 'maybe randomize infohash'
    reuse_torrent_path = 'path/to/existing.torrent'
    announce = 'https://tracker.foo/announce'
    source = 'ASDF'

    kwargs = {
        'content_path': content_path,
        'announce': announce,
        'source': source,
        'torrent_path': torrent_path,
        'exclude': exclude,
        'use_cache': use_cache,
        'reuse_torrent_path': reuse_torrent_path,
        'randomize_infohash': randomize_infohash,
        'piece_size_calculator': piece_size_calculator,
        'piece_size_min_max_calculator': piece_size_min_max_calculator,
        'init_callback': mocks.init_callback,
        'progress_callback': mocks.progress_callback,
    }

    return_value = torrent.create(**kwargs)

    exp_mock_calls = [
        call._get_exclude_regexs(exclude),
        call._get_torrent(
            content_path=content_path,
            exclude=mocks._get_exclude_regexs.return_value,
            announce=announce,
            source=source,
            randomize_infohash=randomize_infohash,
        ),
    ]

    if piece_size_min_max_calculator:
        exp_mock_calls.append(call.piece_size_min_max_calculator(mock_torrent.size))

    if piece_size_calculator:
        exp_mock_calls.append(call.piece_size_calculator(mock_torrent.size))

    exp_mock_calls.append(call.Files(mock_torrent))
    exp_mock_calls.append(call.init_callback(mocks.Files.return_value))

    if not init_cb_cancels and use_cache:
        exp_mock_calls.append(call._find_hashes(
            torrent=mocks._get_torrent.return_value,
            reuse_torrent_path=reuse_torrent_path,
            callback=mocks.progress_callback,
        ))

    if (
            not init_cb_cancels
            and ((not use_cache or (use_cache
                                    and (not find_hashes_return_value
                                         or find_hashes_return_value == torrent.SKIP_SEARCHING)))
                 and not find_hashes_succeeds)
    ):
        exp_mock_calls.append(call._generate_hashes(
            torrent=mocks._get_torrent.return_value,
            callback=mocks.progress_callback,
        ))

    if (
            not init_cb_cancels
            and (not use_cache or (not find_hashes_return_value
                                   or find_hashes_return_value == torrent.SKIP_SEARCHING))
            and (find_hashes_succeeds or not generate_hashes_cancels)
    ):
        exp_mock_calls.extend((
            call._store_generic_torrent(mocks._get_torrent.return_value),
            call._write_torrent_path(mocks._get_torrent.return_value, torrent_path),
        ))
        assert return_value == torrent_path
    else:
        assert return_value is None

    if exp_piece_size is not None:
        assert mock_torrent.piece_size == exp_piece_size
    if exp_piece_size_min is not None:
        assert mock_torrent.piece_size_min == exp_piece_size_min
    if exp_piece_size_max is not None:
        assert mock_torrent.piece_size_max == exp_piece_size_max

    assert mocks.mock_calls == exp_mock_calls


@pytest.mark.parametrize(
    argnames='exclude, exp_result',
    argvalues=(
        (
            [],
            [],
        ),
        (
            ['*.txt', re.compile(r'.*.foo')],
            [re.compile(r'(?s:.*\.txt)\Z', flags=re.IGNORECASE), re.compile(r'.*.foo')],
        ),
        (
            ['*.txt', 123],
            TypeError('Unexpected exclude type: 123'),
        ),
    ),
    ids=lambda v: repr(v),
)
def test_get_exclude_regexs(exclude, exp_result):
    if isinstance(exp_result, Exception):
        with pytest.raises(type(exp_result), match=rf'^{re.escape(str(exp_result))}$'):
            torrent._get_exclude_regexs(exclude)
    else:
        return_value = torrent._get_exclude_regexs(exclude)
        assert return_value == exp_result


@pytest.mark.parametrize(
    argnames='exception, exp_exception',
    argvalues=(
        (None, None),
        (torf.TorfError('foo'), errors.TorrentCreateError('foo')),
    ),
    ids=lambda v: repr(v),
)
def test_get_torrent(exception, exp_exception, mocker):
    Torrent_mock = mocker.patch('torf.Torrent', side_effect=exception)

    kwargs = {
        'content_path': 'path/to/content',
        'exclude': ('foo', 'bar'),
        'announce': 'http://localhost:123/announce',
        'source': 'ASDF',
        'randomize_infohash': 'maybe randomize infohash',
    }

    with patch('time.time', return_value='mock time'):
        if exp_exception:
            with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
                torrent._get_torrent(**kwargs)
        else:
            return_value = torrent._get_torrent(**kwargs)
            assert return_value is Torrent_mock.return_value

    assert Torrent_mock.call_args_list == [call(
        path='path/to/content',
        exclude_regexs=('foo', 'bar'),
        trackers=(('http://localhost:123/announce',),),
        source='ASDF',
        randomize_infohash='maybe randomize infohash',
        private=True,
        created_by=f'{__project_name__} {__version__}',
        creation_date='mock time',
    )]


@pytest.mark.parametrize(
    argnames='exception, exp_exception',
    argvalues=(
        (None, None),
        (torf.TorfError('foo'), errors.TorrentCreateError('foo')),
    ),
    ids=lambda v: repr(v),
)
def test_generate_hashes(exception, exp_exception, mocker):
    CreateTorrentCallback_mock = mocker.patch('upsies.utils.torrent._CreateTorrentCallback')
    callback = Mock()
    mock_torrent = Mock(generate=Mock(side_effect=exception))

    if exp_exception:
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            torrent._generate_hashes(
                torrent=mock_torrent,
                callback=callback,
            )
    else:
        return_value = torrent._generate_hashes(
            torrent=mock_torrent,
            callback=callback,
        )
        assert return_value is CreateTorrentCallback_mock.return_value.return_value

    assert CreateTorrentCallback_mock.call_args_list == [call(callback)]
    assert mock_torrent.generate.call_args_list == [call(
        callback=CreateTorrentCallback_mock.return_value,
        interval=1.0,
    )]


@pytest.mark.parametrize(
    argnames='torrent_files, exception, exp_exception',
    argvalues=(
        (['foo', 'bar'], None, None),
        (['foo', 'bar'], torf.TorfError('argh'), errors.TorrentCreateError('argh')),
        ([], None, None),
        ([], torf.TorfError('argh'), None),
    ),
    ids=lambda v: repr(v),
)
def test_find_hashes(torrent_files, exception, exp_exception, mocker):
    FindTorrentCallback_mock = mocker.patch('upsies.utils.torrent._FindTorrentCallback')
    get_reuse_torrent_paths_mock = mocker.patch('upsies.utils.torrent._get_reuse_torrent_paths')
    callback = Mock()
    mock_torrent = Mock(
        reuse=Mock(side_effect=exception),
        files=torrent_files,
    )

    if exp_exception:
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            torrent._find_hashes(
                torrent=mock_torrent,
                reuse_torrent_path='path/to/existing/torrents/',
                callback=callback,
            )
    else:
        return_value = torrent._find_hashes(
            torrent=mock_torrent,
            reuse_torrent_path='path/to/existing/torrents/',
            callback=callback,
        )
        if torrent_files:
            assert return_value is FindTorrentCallback_mock.return_value.return_value
        else:
            assert return_value is False

    if torrent_files:
        assert FindTorrentCallback_mock.call_args_list == [call(callback)]
        assert mock_torrent.reuse.call_args_list == [call(
            get_reuse_torrent_paths_mock.return_value,
            callback=FindTorrentCallback_mock.return_value,
            interval=1.0,
        )]
    else:
        assert FindTorrentCallback_mock.call_args_list == []
        assert mock_torrent.reuse.call_args_list == []


@pytest.mark.parametrize(
    argnames='reuse_torrent_path, exp_result',
    argvalues=(
        (None, []),
        ((), []),
        ('path/to/existing/torrents', ['path/to/existing/torrents']),
        (('path/to/torrents1', 'path/to/torrents2'), ['path/to/torrents1', 'path/to/torrents2']),
        (123, ValueError('Invalid reuse_torrent_path: 123')),
    ),
    ids=lambda v: repr(v),
)
def test_get_reuse_torrent_paths(reuse_torrent_path, exp_result, mocker):
    get_generic_torrent_path_mock = mocker.patch('upsies.utils.torrent._get_generic_torrent_path')
    mock_torrent = Mock()

    if isinstance(exp_result, Exception):
        with pytest.raises(type(exp_result), match=rf'^{re.escape(str(exp_result))}$'):
            torrent._get_reuse_torrent_paths(mock_torrent, reuse_torrent_path)
    else:
        return_value = torrent._get_reuse_torrent_paths(mock_torrent, reuse_torrent_path)
        assert return_value == [
            get_generic_torrent_path_mock.return_value,
        ] + exp_result

        assert get_generic_torrent_path_mock.call_args_list == [
            call(torrent=mock_torrent, create_directory=False),
        ]


def test_store_generic_torrent(mocker):
    Torrent_mock = mocker.patch('torf.Torrent')
    copy_torrent_info_mock = mocker.patch('upsies.utils.torrent._copy_torrent_info')
    get_generic_torrent_path_mock = mocker.patch('upsies.utils.torrent._get_generic_torrent_path')
    generic_torrent = Torrent_mock.return_value

    with patch('time.time', return_value='mock time'):
        torrent._store_generic_torrent(torrent)

    assert Torrent_mock.call_args_list == [call(
        private=True,
        created_by=f'{__project_name__} {__version__}',
        creation_date='mock time',
        comment='This torrent is used to cache previously hashed pieces.',
    )]
    assert copy_torrent_info_mock.call_args_list == [call(torrent, generic_torrent)]
    assert get_generic_torrent_path_mock.call_args_list == [
        call(generic_torrent, create_directory=True),
    ]
    assert generic_torrent.write.call_args_list == [call(
        get_generic_torrent_path_mock.return_value,
        overwrite=True,
    )]


@pytest.mark.parametrize(
    argnames='exception, exp_exception',
    argvalues=(
        (None, None),
        (torf.TorfError('foo'), errors.TorrentCreateError('foo')),
    ),
    ids=lambda v: repr(v),
)
def test_write_torrent_path(exception, exp_exception, mocker):
    mock_torrent = Mock(write=Mock(side_effect=exception))
    torrent_path = 'path/to/my.torrent'

    if exp_exception:
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            torrent._write_torrent_path(mock_torrent, torrent_path)
    else:
        torrent._write_torrent_path(mock_torrent, torrent_path)

    assert mock_torrent.write.call_args_list == [
        call(torrent_path, overwrite=True),
    ]


@pytest.mark.parametrize(
    argnames='GENERIC_TORRENTS_DIRPATH, create_directory, exception, exp_exception',
    argvalues=(
        ('/generic/torrents', False, errors.ContentError('foo'), None),
        ('/generic/torrents', True, errors.ContentError('foo'), errors.TorrentCreateError('/generic/torrents: foo')),
        ('/generic/torrents', True, None, None),
    ),
    ids=lambda v: repr(v),
)
def test_get_generic_torrent_path(GENERIC_TORRENTS_DIRPATH, create_directory, exception, exp_exception, mocker):
    mocker.patch('upsies.constants.GENERIC_TORRENTS_DIRPATH', GENERIC_TORRENTS_DIRPATH)
    mkdir_mock = mocker.patch('upsies.utils.fs.mkdir', side_effect=exception)
    get_torrent_id_mock = mocker.patch('upsies.utils.torrent._get_torrent_id')
    mock_torrent = Mock()
    mock_torrent.configure_mock(name='My Torrent')

    if exp_exception:
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            torrent._get_generic_torrent_path(mock_torrent, create_directory=create_directory)
    else:
        return_value = torrent._get_generic_torrent_path(mock_torrent, create_directory=create_directory)

    if create_directory:
        assert mkdir_mock.call_args_list == [call(GENERIC_TORRENTS_DIRPATH)]

    if not exp_exception:
        assert get_torrent_id_mock.call_args_list == [call(mock_torrent)]
        assert return_value == os.path.join(
            GENERIC_TORRENTS_DIRPATH,
            f'{mock_torrent.name}.{get_torrent_id_mock.return_value}.torrent',
        )


def test_get_torrent_id_info(mocker):
    class MockFile(str):
        def __new__(cls, string, size):
            self = super().__new__(cls, string)
            self.size = size
            return self

    mock_torrent = Mock()
    mock_torrent.configure_mock(
        name='My Torrent',
        files=[
            MockFile('foo', size=123),
            MockFile('bar', size=234),
        ],
    )

    return_value = torrent._get_torrent_id_info(mock_torrent)
    assert return_value == {
        'name': 'My Torrent',
        'files': (
            ('foo', 123),
            ('bar', 234),
        ),
    }


def test_get_torrent_id(mocker):
    get_torrent_id_info_mock = mocker.patch('upsies.utils.torrent._get_torrent_id_info')
    semantic_hash_mock = mocker.patch('upsies.utils.semantic_hash')

    mock_torrent = Mock()
    return_value = torrent._get_torrent_id(mock_torrent)
    assert return_value is semantic_hash_mock.return_value
    assert semantic_hash_mock.call_args_list == [call(get_torrent_id_info_mock.return_value)]
    assert get_torrent_id_info_mock.call_args_list == [call(mock_torrent)]


@pytest.mark.parametrize(
    argnames='info_a, info_b, exp_info_b',
    argvalues=(
        (
            {'pieces': 'pieces a', 'piece length': 'piece length a', 'name': 'name a', 'length': 'length a', 'foo': 'bar'},
            {},
            {'pieces': 'pieces a', 'piece length': 'piece length a', 'name': 'name a', 'length': 'length a'},
        ),
        (
            {'pieces': 'pieces a', 'piece length': 'piece length a', 'name': 'name a', 'files': 'files a', 'foo': 'bar'},
            {},
            {'pieces': 'pieces a', 'piece length': 'piece length a', 'name': 'name a', 'files': 'files a'},
        ),
        (
            {'pieces': 'pieces a', 'piece length': 'piece length a', 'name': 'name a', 'length': 'length a', 'foo': 'bar'},
            {'pieces': 'pieces b', 'piece length': 'piece length b', 'name': 'name b', 'length': 'length b', 'foo': 'baz'},
            {'pieces': 'pieces a', 'piece length': 'piece length a', 'name': 'name a', 'length': 'length a', 'foo': 'baz'},
        ),
        (
            {'pieces': 'pieces a', 'piece length': 'piece length a', 'name': 'name a', 'files': 'files a', 'foo': 'bar'},
            {'pieces': 'pieces b', 'piece length': 'piece length b', 'name': 'name b', 'files': 'files b', 'foo': 'baz'},
            {'pieces': 'pieces a', 'piece length': 'piece length a', 'name': 'name a', 'files': 'files a', 'foo': 'baz'},
        ),
    ),
    ids=lambda v: str(v),
)
def test_copy_torrent_info(info_a, info_b, exp_info_b):
    info_a_orig = copy.deepcopy(info_a)
    a, b = Mock(metainfo={'info': info_a}), Mock(metainfo={'info': info_b})
    torrent._copy_torrent_info(a, b)
    assert info_b == exp_info_b
    assert info_a == info_a_orig


def test_Files():
    class File(str):
        def __new__(cls, string, size):
            self = super().__new__(cls, string)
            self.size = size
            return self

    print(repr(torrent.Files), id(torrent.Files))

    mock_torrent = Mock(
        filepaths=[
            File('path/to/Top/a', 123),
            File('path/to/Top/b/c', 456),
            File('path/to/Top/b/d', 789),
        ],
        filetree={
            'Top': {
                'a': File('a', 123),
                'b': {
                    'c': File('c', 456),
                    'd': File('d', 789),
                },
            },
        },
        path=SimpleNamespace(parent='path/to'),
    )

    files = torrent.Files(mock_torrent)
    assert files.list == (
        'path/to/Top/a',
        'path/to/Top/b/c',
        'path/to/Top/b/d',
    )
    assert all(isinstance(f, str) for f in files.list)

    assert files.tree == (
        ('path/to/Top', (
            ('a', 123),
            ('b', (
                ('c', 456),
                ('d', 789),
            )),
        )),
    )


@pytest.fixture
def TestCallback():
    class TestCallback(torrent._CallbackBase):
        pass
    return TestCallback

def test_CallbackBase_return_value_attribute(TestCallback):
    return_value = Mock()
    cb = Mock(return_value=return_value)
    mock_cb = TestCallback(cb)
    assert mock_cb.return_value is None
    assert mock_cb('mock progress') is return_value
    assert mock_cb.return_value is return_value
    assert cb.call_args_list == [call('mock progress')]


@pytest.mark.parametrize(
    argnames='time_started, time_current, items_done, items_total, samples',
    argvalues=(
        (100, 160, 200, 300, []),
        (100, 160, 156, 300, [(163, 138), (166, 147)]),
        (100, 112, 60, 300, [(103, 51), (106, 54), (109, 57)]),
        (100, 112, 27, 90, [(100, 27), (101, 27), (102, 27), (103, 27)]),
    ),
    ids=lambda v: repr(v),
)
def test_CallbackBase_calculate_info(time_started, time_current, items_done, items_total, samples, TestCallback, mocker):
    exp_info = {}
    exp_info['percent_done'] = items_done / items_total * 100
    if len(samples) >= 2:
        exp_info['items_per_second'] = (samples[-1][1] - samples[0][1]) / (len(samples) - 1)
    else:
        exp_info['items_per_second'] = 0
    exp_info['items_remaining'] = items_total - items_done
    if exp_info['items_per_second'] > 0:
        exp_info['seconds_remaining'] = datetime.timedelta(seconds=exp_info['items_remaining'] / exp_info['items_per_second'])
    else:
        exp_info['seconds_remaining'] = datetime.timedelta(seconds=0)
    exp_info['seconds_elapsed'] = datetime.timedelta(seconds=time_current - time_started)
    exp_info['seconds_total'] = datetime.timedelta(seconds=(
        time_current - time_started
        + exp_info['seconds_remaining'].total_seconds()
    ))
    exp_info['time_finished'] = datetime.datetime.fromtimestamp(time_current + exp_info['seconds_remaining'].total_seconds())
    exp_info['time_started'] = datetime.datetime.fromtimestamp(time_started)

    add_sample_mock = mocker.patch('upsies.utils.torrent._CallbackBase._add_sample')
    get_average_mock = mocker.patch('upsies.utils.torrent._CallbackBase._get_average')
    get_average_mock.side_effect = lambda diffs, **kw: sum(diffs) / len(diffs)

    cb = TestCallback(Mock())
    cb._time_started = time_started
    cb._progress_samples = samples

    with patch('time.time', return_value=time_current):
        info = cb._calculate_info(items_done, items_total)
    assert info == exp_info

    assert add_sample_mock.call_args_list == [call(time_current, items_done)]

    if samples:
        exp_diffs = [
            b[1] - a[1]
            for a, b in zip(samples[:-1], samples[1:])
        ]
        assert get_average_mock.call_args_list == [call(exp_diffs, weight_factor=1.1)]
    else:
        assert get_average_mock.call_args_list == []


@pytest.mark.parametrize(
    argnames='rounds, interval, items_to_keep',
    argvalues=(
        (15, 6, 2),
        (15, 5, 3),
        (15, 4, 3),
        (15, 3, 4),
        (15, 2, 6),
        (15, 1, 11),
    ),
    ids=lambda v: repr(v),
)
def test_CallbackBase_add_sample(rounds, interval, items_to_keep, TestCallback):
    time_now_pool = iter(range(1000, 2000, interval))
    items_done_pool = iter(range(0, 100, 3))
    exp_progress_samples = []

    cb = TestCallback(Mock())
    assert cb._progress_samples == exp_progress_samples

    for i in range(rounds + 1):
        time_now = next(time_now_pool)
        items_done = next(items_done_pool)
        cb._add_sample(time_now, items_done)

        exp_progress_samples.append((time_now, items_done))
        del exp_progress_samples[:-items_to_keep]

        assert cb._progress_samples == exp_progress_samples


@pytest.mark.parametrize(
    argnames='weight_factor, samples, exp_average',
    argvalues=(
        (1.0, [1, 2, 3, 6, 5, 1], (1 + 2 + 3 + 6 + 5 + 1) / 6),  # sum(samples) = 18 / len(samples) = 6
        (2.0, [3, 2, 6, 15], (
            ((3 * 2) + (2 * 4) + (6 * 8) + (15 * 16))
            / (2 + 4 + 8 + 16)
        )),
        (3.0, [9, 9, 8, 3], (
            ((9 * 3) + (9 * 9) + (8 * 27) + (3 * 81))
            / (3 + 9 + 27 + 81)
        )),
    ),
    ids=lambda v: repr(v),
)
def test_CallbackBase_get_average(weight_factor, samples, exp_average, TestCallback):
    cb = TestCallback(Mock())
    average = cb._get_average(samples, weight_factor)
    assert average == exp_average


def test_CreateTorrentCallback(mocker):
    call_mock = mocker.patch('upsies.utils.torrent._CallbackBase.__call__')
    mocker.patch('upsies.utils.torrent._CreateTorrentCallback._calculate_info', return_value={
        'items_per_second': 100,
        'percent_done': 200,
        'seconds_elapsed': 300,
        'seconds_remaining': 400,
        'seconds_total': 500,
        'time_finished': 600,
        'time_started': 700,
    })

    cb = torrent._CreateTorrentCallback(Mock())
    return_value = cb(
        torrent=Mock(piece_size=123, size=456),
        filepath='path/to/file',
        pieces_done=23,
        pieces_total=42,
    )
    assert return_value is call_mock.return_value
    assert call_mock.call_args_list == [call(torrent.CreateTorrentProgress(
        pieces_done=23,
        pieces_total=42,
        percent_done=200,
        bytes_per_second=types.Bytes(100 * 123),
        piece_size=123,
        total_size=456,
        filepath='path/to/file',
        seconds_elapsed=300,
        seconds_remaining=400,
        seconds_total=500,
        time_finished=600,
        time_started=700,
    ))]
    assert cb._calculate_info.call_args_list == [call(23, 42)]


@pytest.mark.parametrize(
    argnames='exception, exp_exception',
    argvalues=(
        (None, None),
        (OSError('foo'), errors.TorrentCreateError('foo')),
        (torf.TorfError('foo'), errors.TorrentCreateError('foo')),
        (torf.ReadError(errno=errno.EACCES, path='foo'), errors.TorrentCreateError('foo: Permission denied')),
        (torf.ReadError(errno=errno.ENOENT, path='foo'), None),
    ),
    ids=lambda v: repr(v),
)
@pytest.mark.parametrize(
    argnames='status, exp_status',
    argvalues=(
        (True, 'hit'),
        (False, 'miss'),
        (None, 'verifying'),
    ),
)
def test_FindTorrentCallback(exception, exp_exception, status, exp_status, mocker):
    call_mock = mocker.patch('upsies.utils.torrent._CallbackBase.__call__')
    mocker.patch('upsies.utils.torrent._FindTorrentCallback._calculate_info', return_value={
        'items_per_second': 100,
        'percent_done': 200,
        'seconds_elapsed': 300,
        'seconds_remaining': 400,
        'seconds_total': 500,
        'time_finished': 600,
        'time_started': 700,
    })
    cb = torrent._FindTorrentCallback(Mock())
    return_value = cb(
        torrent=Mock(piece_size=123, size=456),
        filepath='path/to/file',
        files_done=23,
        files_total=42,
        status=status,
        exception=exception,
    )
    assert return_value is call_mock.return_value
    assert call_mock.call_args_list == [call(torrent.FindTorrentProgress(
        files_done=23,
        files_total=42,
        percent_done=200,
        files_per_second=100,
        filepath='path/to/file',
        status=exp_status,
        exception=exp_exception,
        seconds_elapsed=300,
        seconds_remaining=400,
        seconds_total=500,
        time_finished=600,
        time_started=700,
    ))]
    assert cb._calculate_info.call_args_list == [call(23, 42)]


def test_filter_files_succeeds(mocker):
    mocks = Mock()
    mocks.attach_mock(
        mocker.patch('upsies.utils.torrent._get_torrent', return_value=Mock(
            filepaths=['this.txt', 'that.jpg'],
        )),
        '_get_torrent',
    )
    mocks.attach_mock(
        mocker.patch('upsies.utils.torrent._get_exclude_regexs'),
        '_get_exclude_regexs',
    )

    content_path = '/path/to/something'
    exclude = ['*.foo', '*bar*']
    filtered_files = torrent.filter_files(content_path, exclude)

    assert filtered_files == ['this.txt', 'that.jpg']
    assert mocks.mock_calls == [
        call._get_exclude_regexs(exclude),
        call._get_torrent(
            content_path=content_path,
            exclude=mocks._get_exclude_regexs.return_value,
            announce='http://dummy.local',
            source=None,
            randomize_infohash=False,
        ),
    ]

def test_filter_files_fails(mocker):
    mocks = Mock()
    mocks.attach_mock(
        mocker.patch('upsies.utils.torrent._get_torrent', side_effect=errors.TorrentCreateError('oh no!')),
        '_get_torrent',
    )
    mocks.attach_mock(
        mocker.patch('upsies.utils.torrent._get_exclude_regexs'),
        '_get_exclude_regexs',
    )

    content_path = '/path/to/something'
    exclude = ['*.foo', '*bar*']
    with pytest.raises(errors.ContentError, match=r'^oh no\!$'):
        torrent.filter_files(content_path, exclude)

    assert mocks.mock_calls == [
        call._get_exclude_regexs(exclude),
        call._get_torrent(
            content_path=content_path,
            exclude=mocks._get_exclude_regexs.return_value,
            announce='http://dummy.local',
            source=None,
            randomize_infohash=False,
        ),
    ]
